from collections import UserString
from pathlib import Path
from typing import Optional


class VersionProxy(UserString):
    """
    Version handling helper that pairs with setuptools-scm.

    This allows for pkg.__version__ to be dynamically retrieved on request by
    way of setuptools-scm.

    This deferred evaluation of the version until it is checked saves time on
    package import.

    This supports the following scenarios:

    1. A git checkout (.git exists)
    2. A git archive / a tarball release from GitHub that includes version
        information in .git_archival.txt.
    3. An existing _version.py generated by setuptools_scm
    4. A fallback in case none of the above match - resulting in a version of
        0.0.unknown
    """
    def __init__(self):
        self._version = None

    def _get_version(self) -> Optional[str]:
        # Checking for directory is faster than failing out of get_version
        repo_root = Path(__file__).resolve().parent.parent
        if (repo_root / ".git").exists() or (repo_root / ".git_archival.txt").exists():
            try:
                # Git checkout
                from setuptools_scm import get_version
                return get_version(root="..", relative_to=__file__)
            except (ImportError, LookupError):
                ...

        # Check this second because it can exist in a git repo if we've
        # done a build at least once.
        try:
            from ._version import version  # noqa: F401
            return version
        except ImportError:
            ...

        return None

    @property
    def data(self) -> str:
        # This is accessed by UserString to allow us to lazily fill in the
        # information
        if self._version is None:
            self._version = self._get_version() or '0.0.unknown'

        return self._version


__version__ = version = VersionProxy()



"""
This module defines the ``happi`` command line interface.
"""
from __future__ import annotations

import ast
import dataclasses
import fnmatch
import importlib
import inspect
import json
import logging
import os
# on import allows arrow key navigation in prompt
import readline  # noqa
import subprocess
import sys
import time
from contextlib import contextmanager
from cProfile import Profile
from pathlib import Path
from typing import Optional

import click
import coloredlogs
import platformdirs
import prettytable

import happi
from happi.errors import SearchError

from .audit import audit as run_audit
from .audit import (checks, find_unfilled_mandatory_info,
                    find_unfilled_optional_info)
from .prompt import prompt_for_entry, transfer_container
from .utils import is_a_range, is_number, is_valid_identifier_not_keyword

logger = logging.getLogger(__name__)

version_msg = f'Happi: Version {happi.__version__} from {happi.__file__}'
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


def get_happi_client_from_config(path):
    # gather client
    try:
        client = happi.client.Client.from_config(path)
        logger.debug("Happi client: %r" % client)
        # insert items into context to be passed to subcommands
        # User objects must be assigned to ctx.obj, which will be passed
        # through to new context objects
        return client
    except OSError:
        logger.debug("Happi configuration not found.")


@click.group(
    help=('The happi command-line interface, used to view and manipulate '
          'device databases'),
    context_settings=CONTEXT_SETTINGS
)
@click.option('--path', type=click.Path(exists=True),
              help='Provide the path to happi configuration file. '
                   'Will default to the file stored in the HAPPI_CFG '
                   'environment variable.')
@click.option('--verbose', '-v', is_flag=True,
              help='Show the debug logging stream.')
@click.version_option(None, '--version', '-V', message=version_msg)
@click.pass_context
def happi_cli(ctx, path, verbose):
    """Happi command line tool"""
    # Logging Level handling
    if verbose:
        shown_logger = logging.getLogger()
        level = "DEBUG"
    else:
        shown_logger = logging.getLogger('happi')
        level = "INFO"
    coloredlogs.install(level=level, logger=shown_logger,
                        fmt='[%(asctime)s] - %(levelname)s -  %(message)s')
    logger.debug("Set logging level of %r to %r", shown_logger.name, level)

    ctx.obj = path

    # Cleanup tasks related to loaded devices
    @ctx.call_on_close
    def device_cleanup():
        pyepics_cleanup()
        ophyd_cleanup()


@happi_cli.command()
@click.option('--show_json', '--json', '-j', is_flag=True,
              help='Show results in JSON format.')
@click.option('--names', is_flag=True,
              help='Return results as whitespace-separated names.')
@click.option('--glob/--regex', 'use_glob', default=True,
              help='use glob style (default) or regex style search terms. '
              r'Regex requires backslashes to be escaped (eg. at\\d.\\d)')
@click.argument('search_criteria', nargs=-1)
@click.pass_context
def search(
    ctx: click.Context,
    show_json: bool,
    names: bool,
    use_glob: bool,
    search_criteria: list[str]
):
    """
    Search the happi database.  SEARCH_CRITERIA take the form: field=value.
    If 'field=' is omitted, it will assumed to be 'name'.
    You may include as many search criteria as you like; these will
    be combined with ANDs.
    """
    logger.debug("We're in the search block")

    final_results = search_parser(
        client=get_happi_client_from_config(ctx.obj),
        use_glob=use_glob,
        search_criteria=search_criteria,
    )
    if not final_results:
        return []

    # Final processing for output
    if show_json:
        json.dump([dict(res.item) for res in final_results], indent=2,
                  fp=sys.stdout)
    elif names:
        out = " ".join([res.item.name for res in final_results])
        click.echo(out)
    else:
        for res in final_results:
            res.item.show_info()

    return final_results


def search_parser(
    client: happi.Client,
    use_glob: bool,
    search_criteria: list[str],
) -> list[happi.SearchResult]:
    """
    Parse the user's search criteria and return the search results.

    ``search_criteria`` must be a list of key=value strings.
    If key is omitted, it will be assumed to be "name".

    Parameters
    ----------
    client : Client
        The happi client that we'll be doing searches in.
    use_glob : bool
        True if we're using glob matching, False if we're using
        regex matching.
    search_criteria : list of str
        The user's search selection from the command line.
    """
    # Get search criteria into dictionary for use by client
    client_args = {}
    range_list = []
    regex_list = []
    range_found = False

    with client.retain_cache_context():
        for user_arg in search_criteria:
            if '=' in user_arg:
                criteria, value = user_arg.split('=', 1)
            else:
                criteria = 'name'
                value = user_arg
            if criteria in client_args:
                raise click.ClickException(
                    f"Received duplicate search criteria {criteria}={value!r} "
                    f"(was {client_args[criteria]!r})"
                )

            if is_a_range(value):
                start, stop = value.split(',')
                start = float(start)
                stop = float(stop)
                if start < stop:
                    new_range_list = client.search_range(criteria, start, stop)
                else:
                    raise click.ClickException('Invalid range, make sure start < stop')

                if not range_found:
                    # if first range, just replace
                    range_found = True
                    range_list = new_range_list
                else:
                    # subsequent ranges, only take intersection
                    range_list = set(new_range_list) & set(range_list)

                if not range_list:
                    # we have searched via a range query.  At this point
                    # no matches, or intesection is empty. abort early
                    logger.error("No items found")
                    return []

                continue

            elif is_number(value):
                if float(value) == int(float(value)):
                    # value is an int, allow the float version (optional .0)
                    logger.debug(f'looking for int value: {value}')
                    value = f'^{int(float(value))}(\\.0+$)?$'

                    # don't translate from glob
                    client_args[criteria] = value
                    continue
                else:
                    value = str(float(value))
            else:
                logger.debug('Value %s interpreted as string', value)

            if use_glob:
                client_args[criteria] = fnmatch.translate(value)
            else:  # already using regex
                client_args[criteria] = value

        regex_list = client.search_regex(**client_args)

    # Gather final results
    final_results = []
    if regex_list and not range_list:
        # only matched with one search_regex()
        final_results = regex_list
    elif range_list and not regex_list:
        # only matched with search_range()
        final_results = range_list
    elif range_list and regex_list:
        # find the intersection between regex_list and range_list
        final_results = set(range_list) & set(regex_list)
    else:
        logger.debug('No regex or range items found')

    if not final_results:
        logger.error('No items found')
    return final_results


@happi_cli.command()
@click.option('--clone', default='',
              help='Copy the fields from an existing container. '
              'Provide the name of the item to clone.')
@click.pass_context
def add(ctx, clone: str):
    """Add new entries or copy existing entries."""
    logger.debug(f'Starting interactive add, {clone}')
    # retrieve client
    client = get_happi_client_from_config(ctx.obj)

    registry = happi.containers.registry
    if clone:
        clone_source = client.find_item(name=clone)
        # Must use the same container if cloning
        response = registry.entry_for_class(clone_source.__class__)
    else:
        clone_source = None
        # Keep Device at registry for backwards compatibility but filter
        # it out of new item options
        options = os.linesep.join(
            [k for k, _ in registry.items() if k != "Device"]
        )
        ctnr_prompt = (
            'Please select a container, or press enter for generic '
            f'Ophyd Device container: {os.linesep}{options}'
            f'{os.linesep}{os.linesep}Selection'
        )
        logger.debug(ctnr_prompt)
        response = click.prompt(ctnr_prompt, default='OphydItem')
        if response and response not in registry:
            raise click.ClickException(f'Invalid item container {response}')

    container = registry[response]
    logger.debug(f'Contaner selected: {container.__name__}')

    # Collect values for each field
    kwargs = {}
    for info in container.entry_info:
        item_value = prompt_for_entry(info, clone_source=clone_source)
        click.echo(f'Selecting value: {item_value}')

        try:
            info.enforce_value(item_value)
            kwargs[info.key] = item_value
        except Exception:
            logger.info(f'Invalid value {item_value}')

    item = client.create_item(container, **kwargs)
    item.show_info()
    if click.confirm('Please confirm the item info is correct'):
        logger.info('Adding item')
        item.save()
    else:
        logger.info('Aborting')


@happi_cli.command()
@click.argument('name', type=str)
@click.pass_context
def copy(ctx, name):
    """
    Copy the item NAME.

    Simply wraps ``happi add --clone``
    """
    ctx.invoke(add, clone=name)


@happi_cli.command()
@click.argument('name', type=str)
@click.pass_context
def delete(ctx, name: str):
    """
    Delete an existing entry.  Only accepts exact names
    """
    client = get_happi_client_from_config(ctx.obj)
    try:
        item = client.find_item(name=name)
    except SearchError as e:
        raise click.ClickException(f'Could not find item ({name}): {e}')

    item.show_info()
    if click.confirm('Are you sure you want to delete this entry? \n'
                     'Remember you can mark an entry as inactive with \n'
                     '"happi edit my_device_name active=false"'):
        logger.info('Deleting item')
        client.remove_item(item)
        click.echo(f'Entry {name} removed')
    else:
        logger.info('Aborting')


@happi_cli.command()
@click.argument('name')
@click.argument('edits', nargs=-1, type=str)
@click.pass_context
def edit(ctx, name: str, edits: list[str]):
    """
    Change an existing entry.

    Applies EDITS of the form: field=value to the item of name NAME.
    """
    # retrieve client
    client = get_happi_client_from_config(ctx.obj)

    logger.debug('Starting edit block')
    try:
        item = client.find_item(name=name)
    except SearchError as e:
        raise click.ClickException(f'Could not find item {name!r}: {e}')

    if len(edits) < 1:
        raise click.ClickException('No edits provided')

    for edit in edits:
        field, value = edit.split('=', 1)
        # validate field
        try:
            field = is_valid_identifier_not_keyword(field)
        except ValueError:
            raise click.ClickException(f'field {field!r} is not a valid identifier')

        # validate field depending on enforce type
        if item._info_attrs[field].enforce in (dict, list):
            # try interpreting with ast
            try:
                value = ast.literal_eval(value)
            except (ValueError, SyntaxError) as e:
                raise click.ClickException(f'Could not interpret edit ({edit}), {e}\n'
                                           'Remember to escape your quotes '
                                           '(e.g. {\\"key\\":\\"value\\"})')
        else:
            try:
                value = item._info_attrs[field].enforce_value(value)
            except ValueError:
                raise click.ClickException(f'Error enforcing type in field {field!r} for value: {value!r}')

        # set field
        try:
            getattr(item, field)
            logger.info(f'Setting {name}.{field}: {value}')
            setattr(item, field, value)
        except Exception as e:
            raise click.ClickException(f'Could not set {name}.{field} to {value!r}: {e}')

    item.show_info()
    if click.confirm('Please confirm the item info is correct'):
        logger.info('Editing item')
        item.save()
    else:
        logger.info('Aborting')


@happi_cli.command()
@click.option('--glob/--regex', 'use_glob', default=True,
              help='use glob style (default) or regex style search terms. '
              r'Regex requires backslashes to be escaped (eg. at\\d.\\d)')
@click.argument('search_criteria', nargs=-1)
@click.pass_context
def load(
    ctx: click.Context,
    use_glob: bool,
    search_criteria: list[str]
):
    results = set()

    for item in search_criteria:
        final_results = search_parser(
            client=get_happi_client_from_config(ctx.obj),
            use_glob=use_glob,
            search_criteria=[item],
        )
        if not final_results:
            print('%s was not found.' % item)
        else:
            for res in final_results:
                results.add(res['name'])

    # Open IPython terminal with RESULTS loaded

    logger.debug('Starting load block')
    # retrieve client
    client = get_happi_client_from_config(ctx.obj)

    devices = {}
    names = " ".join(results)
    names = names.split()

    if len(names) < 1:
        raise click.BadArgumentUsage('No item names given')
    logger.info(f'Creating shell with devices {results}')

    for name in names:
        try:
            devices[name] = client.load_device(name=name)
        except SearchError as e:
            raise click.ClickException(f'Could not load device {name!r}: {e}')

    try:
        from IPython import start_ipython  # noqa
        start_ipython(argv=['--quick'], user_ns=devices)
    except ImportError:
        # Fall back to normal Python REPL if IPython is not available
        import code
        vars = globals().copy()
        vars.update(devices)
        shell = code.InteractiveConsole(vars)
        shell.interact()


# TODO: Figure out how to deal with json and click.  list of args doesn't
# translate exactly
@happi_cli.command()
@click.argument("json_data", nargs=-1)
@click.pass_context
def update(ctx, json_data: str):
    """
    Update happi db with JSON_DATA payload.

    To use, either use command substitution:

        $ happi update $(cat my.json)

    Or pipe the JSON payload with `-` as an argument:

        $ cat my.json | happi update -

    JSON payloads should be a list of items (dictionaries), with at least the
    "_id" and "type" keys.  eg:

    \b
        [{
            "_id": "my_device",
            <...>
            "type": "mydevicelibrary.MyDevice"
        }]

    Or a valid happi json database.  eg:

    \b
        {
            "my_device": {
                "_id": "my_device",
                <...>
                "type": "mydevicelibrary.MyDevice"
            }
        }
    """
    # retrieve client
    client = get_happi_client_from_config(ctx.obj)
    if len(json_data) < 1:
        raise click.BadArgumentUsage('No json data given')

    # parse input
    input_ = " ".join(json_data).strip()
    if input_ == "-":
        items_input = json.load(sys.stdin)
    else:
        items_input = json.loads(input_)

    # insert
    if isinstance(items_input, dict):
        items_input = items_input.values()

    for item in items_input:
        item = client.create_item(item["type"], **item)
        exists = item["_id"] in [c["_id"] for c in client.all_items]
        client._store(item, insert=not exists)


@happi_cli.command(name='container-registry')
def container_registry():
    """Print container registry"""
    pt = prettytable.PrettyTable()
    pt.field_names = ["Container Name", "Container Class", "Object Class"]
    pt.align = "l"
    for type_, class_, in happi.containers.registry.items():
        pt.add_row([type_,
                    f'{class_.__module__}.{class_.__name__}',
                    class_.device_class.default])
    click.echo(pt)


@happi_cli.command()
@click.pass_context
@click.argument("name", type=str, nargs=1)
@click.argument("target", type=str, nargs=1)
def transfer(ctx, name: str, target: str):
    """
    Change the container of an item.

    Transfers item (NAME) to a new container (TARGET)
    """
    logger.debug('Starting transfer block')
    # retrive client
    client = get_happi_client_from_config(ctx.obj)
    # verify name and target both exist and are valid
    try:
        item = client.find_item(name=name)
    except SearchError as e:
        raise click.ClickException(str(e))
    registry = happi.containers.registry
    # This is slow if dictionary is large
    target_match = [k for k, _ in registry.items() if target in k]
    if len(target_match) > 1 and name in target_match:
        target_match = [name]
    elif len(target_match) != 1:
        raise click.ClickException(f'Target container name ({target}) not specific enough. Possible matches: {target_match}')

    target = happi.containers.registry._registry[target_match[0]]
    # transfer item and prompt for fixes
    transfer_container(client, item, target)


benchmark_sort_keys = [
    'name',
    'avg_time',
    'iterations',
    'tot_time',
    'max_time',
    'import_time',
]


@happi_cli.command()
@click.pass_context
@click.option("-d", "--duration", type=float, default=0,
              help="Specify how long in seconds to spend per device.")
@click.option("-i", "--iterations", type=int, default=1,
              help="Specify the number of times to instantiate each device.")
@click.option("-w", "--wait-connected", is_flag=True,
              help="Wait for the devices to be connected.")
@click.option("-t", "--tracebacks", is_flag=True,
              help="Show tracebacks from failing device loads.")
@click.option("-s", "--sort-key", type=str, default="avg_time",
              help=("Sort the output table. Valid options are "
                    f"{', '.join(benchmark_sort_keys)}"))
@click.option('--glob/--regex', 'use_glob', default=True,
              help='Use glob style (default) or regex style search terms. '
              r'Regex requires backslashes to be escaped (eg. at\\d.\\d)')
@click.argument('search_criteria', nargs=-1)
def benchmark(
    ctx,
    duration: float,
    iterations: int,
    wait_connected: bool,
    tracebacks: bool,
    sort_key: str,
    use_glob: bool,
    search_criteria: list[str],
):
    """
    Compare happi device startup times.

    This will generate a table that shows you how long each device took
    to instantiate.

    Repeats for at least the (-d, --duration) arg (default = 0 seconds)
    and for at least the number of the (-i, --iterations) arg (default = 1
    iteration), showing stats and averages.

    By default we time only the duration of __init__, but you can also
    (wait_connected) to see the full time until the device is fully ready
    to go, presuming the device has a wait_for_connection method.

    Search terms are standard as in the same search terms as the search
    cli function. A blank search term means to load all the devices.
    """
    logger.debug('Starting benchmark block')
    client: happi.Client = get_happi_client_from_config(ctx.obj)
    full_stats = []
    logger.info('Collecting happi items...')
    start = time.monotonic()
    if search_criteria:
        items = search_parser(
            client=client,
            use_glob=use_glob,
            search_criteria=search_criteria,
        )
    else:
        # All the items
        items = client.search()
    logger.info(f'Done, took {time.monotonic() - start} s')
    for result in items:
        logger.info(f'Running benchmark on {result["name"]}')
        try:
            stats = Stats.from_search_result(
                result=result,
                duration=duration,
                iterations=iterations,
                wait_connected=wait_connected,
            )
        except Exception:
            logger.error(
                f'Error running benchmark on {result["name"]}',
                exc_info=tracebacks,
            )
        else:
            full_stats.append(stats)
    table = prettytable.PrettyTable()
    table.field_names = benchmark_sort_keys
    if sort_key not in table.field_names:
        logger.warning(f'Sort key {sort_key} invalid, reverting to avg_time')
        sort_key = 'avg_time'
    for stats in sorted(
        full_stats,
        key=lambda x: getattr(x, sort_key),
        reverse=True,
    ):
        table.add_row([getattr(stats, key) for key in benchmark_sort_keys])
    print('Benchmark output:')
    print(table)
    print('Benchmark completed successfully')


@dataclasses.dataclass
class Stats:
    """
    Collect and hold results from benchmark runs.
    """
    name: str
    avg_time: float
    iterations: int
    tot_time: float
    max_time: float
    import_time: float

    @classmethod
    def from_search_result(
        cls,
        result: happi.SearchResult,
        duration: float,
        iterations: int,
        wait_connected: bool,
    ) -> Stats:
        """
        Create an object using a search result and store benchmarking info.
        """
        logger.debug(f'Checking stats for {result["name"]}')
        if not duration and not iterations:
            return Stats(
                name=result["name"],
                avg_time=0,
                iterations=0,
                tot_time=0,
                max_time=0,
                import_time=0,
            )
        raw_stats: list[float] = []
        import_time = cls.import_benchmark(result)
        counter = 0
        start = time.monotonic()
        while counter < iterations or time.monotonic() - start < duration:
            raw_stats.append(
                cls.run_one_benchmark(
                    result=result,
                    wait_connected=wait_connected
                )
            )
            counter += 1
        return Stats(
            name=result["name"],
            avg_time=sum(raw_stats) / len(raw_stats),
            iterations=len(raw_stats),
            tot_time=sum(raw_stats),
            max_time=max(raw_stats),
            import_time=import_time,
        )

    @staticmethod
    def import_benchmark(result: happi.SearchResult) -> float:
        """
        Check only the module import in isolation.
        """
        start = time.monotonic()
        happi.loader.import_class(result.item.device_class)
        return time.monotonic() - start

    @staticmethod
    def run_one_benchmark(
        result: happi.SearchResult,
        wait_connected: bool,
    ) -> float:
        """
        Create one object and time it.
        """
        start = time.monotonic()
        device = result.get(use_cache=False)
        if wait_connected:
            try:
                device.wait_for_connection(timeout=10.0)
            except AttributeError:
                logger.warning(
                    f'{result["name"]} does not have wait_for_connection.'
                )
            except TimeoutError:
                logger.warning(
                    'Timeout after 10s while waiting for connection of '
                    f'{result["name"]}',
                )
            except Exception:
                logger.warning(
                    'Unknown exception while waiting for connection of '
                    f'{result["name"]}',
                )
        return time.monotonic() - start


@happi_cli.command()
@click.pass_context
@click.option('-d', '--database', 'profile_database', is_flag=True,
              help='Profile the database loading.')
@click.option('-i', '--import', 'profile_import', is_flag=True,
              help='Profile the module importing.')
@click.option('-o', '--object', 'profile_object', is_flag=True,
              help='Profile the object instantiation.')
@click.option('-a', '--all', 'profile_all', is_flag=True,
              help='Shortcut for enabling all profile stages.')
@click.option('-p', '--profiler', default='auto',
              help='Select which profiler to use.')
@click.option('--glob/--regex', 'use_glob', default=True,
              help='Use glob style (default) or regex style search terms. '
              r'Regex requires backslashes to be escaped (eg. at\\d.\\d)')
@click.argument('search_criteria', nargs=-1)
def profile(
    ctx,
    profile_database: bool,
    profile_import: bool,
    profile_object: bool,
    profile_all: bool,
    profiler: str,
    use_glob: bool,
    search_criteria: list[str],
):
    """
    Per-function startup speed diagnostic.

    This will go through the happi loading process and show
    information about the execution time of all the
    functions called during the process.

    Contains options for picking which devices to check and which
    part of the loading process to profile. You can choose to
    profile the happi database loading (-d, --database), the
    class imports (-i, --import), the object instantiation
    (-o, --object), or all of the above (-a, --all).

    By default this will use whichever profiler you have installed,
    but this can also be overriden with the (-p, --profiler) option.
    The priority order is, first, the pcdsutils line_profiler wrapper
    (--profiler pcdsutils), and second, the built-in cProfile module
    (--profiler cprofile). More options may be added later.

    Search terms are standard as in the same search terms as the search
    cli function. A blank search term means to load all the devices.
    """
    logger.debug('Starting profile block')
    if profiler not in ('auto', 'pcdsutils', 'cprofile'):
        raise RuntimeError(f'Invalid profiler selection {profiler}')

    client: happi.Client = get_happi_client_from_config(ctx.obj)
    if profile_all:
        profile_database = True
        profile_import = True
        profile_object = True
    if not any((profile_database, profile_import, profile_object)):
        raise RuntimeError('No profile options selected!')
    if profiler == 'auto':
        try:
            import pcdsutils.profile
            if pcdsutils.profile.has_line_profiler:
                profiler = 'pcdsutils'
            else:
                profiler = 'cprofile'
        except ImportError:
            profiler = 'cprofile'
    if profiler == 'pcdsutils':
        from pcdsutils.profile import profiler_context
        context_profiler = None
    elif profiler == 'cprofile':
        context_profiler = Profile()

        @contextmanager
        def profiler_context(*args, **kwargs):
            context_profiler.enable()
            yield
            context_profiler.disable()

    @contextmanager
    def null_context(*args, **kwargs):
        yield

    def output_profile():
        # Call at the end: let's output results to stdout
        if profiler == 'pcdsutils':
            from pcdsutils.profile import print_results
            print_results()
        elif profiler == 'cprofile':
            context_profiler.print_stats(sort='cumulative')
        print('Profile completed successfully')

    # Profile stage 1: searching the happi database
    logger.info('Searching the happi database')
    if profile_database:
        db_context = profiler_context
    else:
        db_context = null_context
    start = time.monotonic()
    with db_context(
        module_names=['happi'],
        use_global_profiler=True,
        output_now=False,
    ):
        if search_criteria:
            items = search_parser(
                client=client,
                use_glob=use_glob,
                search_criteria=search_criteria,
            )
        else:
            # All the items
            items = client.search()
    logger.info(
        f'Searched the happi database in {time.monotonic() - start} s'
    )

    if not any((profile_import, profile_object)):
        return output_profile()

    # Profile stage 2: import the device classes
    logger.info('Importing the device classes')
    if profile_import:
        import_context = profiler_context
    else:
        import_context = null_context

    classes = set()
    start = time.monotonic()
    with import_context(
        module_names=['happi'],
        use_global_profiler=True,
        output_now=False,
    ):
        for search_result in items:
            try:
                cls = happi.loader.import_class(
                    search_result.item.device_class,
                )
                classes.add(cls)
            except ImportError:
                logger.warning(
                    f'Failed to import {search_result.item.device_class}'
                )
    logger.info(
        f'Imported the device classes in {time.monotonic() - start} s'
    )

    if not profile_object:
        return output_profile()

    # Check which modules to focus on for line profiler
    module_names = {'happi'}
    for instance_class in classes:
        try:
            parents = instance_class.mro()
        except AttributeError:
            # E.g. we have a function
            parents = [instance_class]
        for parent_class in parents:
            module = parent_class.__module__
            module_names.add(module.split('.')[0])
    # Add imported ophyd control layers
    for cl in ('epics', 'caproto'):
        if cl in sys.modules:
            module_names.add(cl)

    # Profile stage 3: create the device classes
    logger.info('Creating the device classes')
    if profile_object:
        object_context = profiler_context
    else:
        object_context = null_context
    start = time.monotonic()
    with object_context(
        module_names=module_names,
        use_global_profiler=True,
        output_now=False,
    ):
        for search_result in items:
            try:
                search_result.get(use_cache=False)
            except Exception:
                logger.warning(
                    f'Failed to create {search_result["name"]}'
                )
    logger.info(
        f'Created the device classes in {time.monotonic() - start} s'
    )
    return output_profile()


def ophyd_cleanup():
    """
    Clean up ophyd - avoid teardown errors by stopping callbacks.

    If this is not run, ophyd callbacks continue to run and can cause
    terminal spam and segfaults.
    """
    if 'ophyd' in sys.modules:
        import ophyd
        dispatcher = ophyd.cl.get_dispatcher()
        if dispatcher is not None:
            dispatcher.stop()


def pyepics_cleanup():
    """
    Clean up pyepics - avoid teardown errors by stopping callbacks.

    If this is not run, pyepics callbacks continue to run and if they throw
    exceptions they will create terminal spam.

    Run this before ophyd_cleanup to prevent race conditions where pyepics
    is trying to call ophyd things that have been torn down.
    """
    if 'epics' in sys.modules:
        from epics import ca

        # Prevent new callbacks from being set up
        def no_create_channel(*args, **kwargs):
            ...

        ca.create_channel = no_create_channel

        # Remove references to existing callbacks
        for context_cache in ca._cache.values():
            for cache_item in context_cache.values():
                try:
                    cache_item.callbacks.clear()
                    cache_item.access_event_callback.clear()
                except AttributeError:
                    print(cache_item)


@happi_cli.command()
@click.pass_context
@click.option('-f', '--file', 'ext_file',
              type=click.Path(exists=True, dir_okay=False, resolve_path=True),
              help='File to import additional checks from.')
@click.option('-l', '--list', 'list_checks', is_flag=True,
              help='List the available validation checks')
@click.option('-c', '--check', 'check_choices', multiple=True, default=[],
              help='Name of the check to include.  '
                   'Can also provide a substring')
@click.option('-d', '--details', 'details',
              help='Show the details of the specified audit function(s)')
@click.option('--glob/--regex', 'use_glob', default=True,
              help='Use glob style (default) or regex style search terms. '
              r'Regex requires backslashes to be escaped (eg. at\\d.\\d)')
@click.option('--names', '-n', 'names_only', is_flag=True,
              help='Only display names of failed entries')
@click.option('--json', '-j', 'show_json', is_flag=True,
              help='output results in json format')
@click.argument('search_criteria', nargs=-1)
def audit(
    ctx,
    list_checks: bool,
    ext_file: str | None,
    check_choices: list[str],
    details: str,
    use_glob: bool,
    names_only: bool,
    show_json: bool,
    search_criteria: tuple[str, ...]
):
    """
    Audit the current happi database.

    Runs checks on the devices matching the provided SEARCH_CRITERIA.
    Checks are simple functions that raise exceptions on failure,
    whether naturally or via assert calls.  These functions take a single
    happi.SearchResult as an positional argument and returns None if
    successful.

    To import additional checks, provide a file with your check function
    and a list named ``checks`` containing the desired functions.
    """
    logger.debug('Starting audit block')

    # if a file is provided, make its functions available
    if ext_file:
        fp = Path(ext_file)
        sys.path.insert(1, str(fp.parent))
        ext_module = importlib.import_module(fp.stem)
        ext_checks = getattr(ext_module, 'checks')
        checks.extend(ext_checks)

    # List checks subcommand
    if list_checks:
        check_pt = prettytable.PrettyTable(field_names=['name',
                                                        'description'])
        check_pt.hrules = prettytable.ALL
        check_pt.align['description'] = 'l'
        for chk in checks:
            check_pt.add_row([chk.__name__, inspect.cleandoc(chk.__doc__)])
        print(check_pt)
        return

    if details:
        check_fns = [fn for fn in checks if details in fn.__name__]
        for fn in check_fns:
            click.echo(inspect.getsource(fn))

        return

    # gather selected checks
    if check_choices:
        check_list = []
        for check_name in check_choices:
            # check if provided check name is a substring of any checks
            matches = [fn for fn in checks if check_name in fn.__name__]
            if len(matches) != 1:
                raise click.BadParameter(
                    f'provided check name ({check_name}) must match only'
                    f'one check.  Matches: ({[ch.__name__ for ch in matches]})'
                )
            check_list.append(matches[0])
    else:
        # take all checks
        check_list = checks

    client: Optional[happi.Client] = get_happi_client_from_config(ctx.obj)
    if client is None:
        raise click.ClickException("Failed to create happi client")

    results = search_parser(client, use_glob, list(search_criteria))
    logger.info(f'found {len(results)} items to verify')
    logger.info(f'running checks: {[f.__name__ for f in check_list]}')

    final_dict = run_audit(
        results,
        redirect=True,
        verbose=not (names_only or show_json),
        check_list=check_list,
    )
    test_results = final_dict["test_results"]

    # print outs
    if names_only:
        click.echo(' '.join(final_dict["failed_names"]))
    elif show_json:
        json.dump(final_dict, indent=2, fp=sys.stdout)
    else:
        pt = prettytable.PrettyTable(field_names=['name', 'check', 'error'])
        pt.align['error'] = 'l'
        last_name = ''
        for name, success, check, msg in zip(test_results['name'],
                                             test_results['success'],
                                             test_results['check'],
                                             test_results['msg']):
            if not success:
                if name != last_name:
                    if last_name != '':  # initial condition
                        pt.add_row(['', '', ''], divider=True)
                    pt.add_row([name, check, msg])
                else:
                    pt.add_row(['', check, msg])
                last_name = name

        try:
            term_width = os.get_terminal_size()[0]
            pt._max_width = {'error': max(60, term_width - 40)}
        except OSError:
            # non-interactive mode (piping results). default max width
            pt._max_width = {'error': 100}
            pass

        if len(pt.rows) > 0:
            click.echo(pt)

        num_failures = final_dict["failures"]
        click.echo(f'# devices failed: {num_failures} / {len(results)}')


@happi_cli.command()
@click.pass_context
@click.option('--fix-optional/--ignore-optional', 'fix_optional', default=False,
              help='Also prompt for user input on optional information')
@click.option('--glob/--regex', 'use_glob', default=True,
              help='Use glob (default) or regex style search terms. '
              'Only relevant if search_criteria are provided.')
@click.argument('search_criteria', nargs=-1)
def repair(
    ctx,
    fix_optional: bool,
    use_glob: bool,
    search_criteria: tuple[str, ...]
):
    """
    Repair the database.

    Repairs all entries matching SEARCH_CRITERIA, repairs entire database otherwise.

    Entries that don't get any fields changed will not get saved (i.e. their last-edit times will not change).
    """
    logger.debug('starting repair block')

    client: happi.Client = get_happi_client_from_config(ctx.obj)
    if search_criteria:
        results = search_parser(client, use_glob, search_criteria)
    else:
        # run repair on all items
        results = search_parser(client, True, '*')

    for res in results:
        # don't save if changes not made
        changes_made = False

        # fix mandatory info with missing defaults
        req_info = find_unfilled_mandatory_info(res)

        if fix_optional:
            req_info.extend(find_unfilled_optional_info(res))

        res_id = res['_id']

        # fix each mandatory field
        logger.info(f'repairing ({res_id})...')
        for req_field in req_info:
            info = res.item._info_attrs[req_field]
            req_value = prompt_for_entry(info)
            curr_value = getattr(res.item, req_field)

            if curr_value != req_value:
                info.enforce_value(req_value)
                setattr(res.item, req_field, req_value)
                changes_made = True

        # check name and id parity
        if res['name'] != res_id:
            # set name to match id
            res.item.name = res_id
            changes_made = True

        if not changes_made:
            logger.info(f'no actual changes made during repair of {res_id}, not saving...')
            continue

        # re-save after creating container
        try:
            # will save optional data with defaults
            # optional data without defaults saved as null
            res.item.save()
        except KeyboardInterrupt:
            # Finish the current save if interrupted
            logger.warning('caught keyboard interrupt, finishing')
            res.item.save()
            break


@click.group(help="Commands related to the happi config file.")
def config():
    pass


@config.command(name="edit")
def edit_config():
    """Open happi configuration file for editing."""
    config_filepath = happi.client.Client.find_config()
    if sys.platform.startswith("win32"):
        import shutil
        editor = shutil.which(os.environ.get("EDITOR", "notepad.exe"))
        subprocess.run([editor, config_filepath])
    else:
        subprocess.run([os.environ.get("EDITOR", "vi"), config_filepath])


@config.command()
@click.option("--overwrite/--no-overwrite", "overwrite", default=False,
              help="Overwrite existing config.")
@click.option("--backend", type=click.Choice(["json"], case_sensitive=False), default="json")
def init(overwrite, backend):
    """Create configuration file with default options."""

    # find config_filepath
    try:
        config_filepath = Path(happi.client.Client.find_config())
    except OSError:
        config_filepath = Path(platformdirs.user_config_dir("happi")) / "happi.cfg"
    else:
        if not overwrite:
            click.echo("Found existing config file at:")
            click.echo(f"  {config_filepath}")
            click.echo("Stopping! Use --overwrite to destroy this config file.")
            return
    click.echo("Creating new config file at:")
    click.echo(f"  {config_filepath}")

    # find database_filepath
    database_filepath = Path(platformdirs.user_data_dir("happi")) / "db.json"
    if database_filepath.exists():
        click.echo("Using existing database file at:")
    else:
        click.echo("Creating new database file at:")

    click.echo(f"  {database_filepath}")

    # create config file
    config_filepath.parent.mkdir(parents=True, exist_ok=True)
    with open(config_filepath, "w") as f:
        f.write("[DEFAULT]\n")
        f.write(f"path={database_filepath}\n")

    # create database file
    database_filepath.parent.mkdir(parents=True, exist_ok=True)
    database_filepath.touch(exist_ok=True)

    click.echo("Done!")


@config.command()
def show():
    """Show configuration file in current state."""
    config_filepath = Path(happi.client.Client.find_config())
    click.echo(f"File: {config_filepath}")

    def draw_line():
        try:
            click.echo("-"*os.get_terminal_size()[0])
        except OSError:
            # non-interactive mode (piping results). No max width
            click.echo("-"*79)

    draw_line()
    with open(config_filepath, "r") as f:
        for line in f:
            click.echo(line.strip())
    draw_line()


def main():
    """Execute the ``happi_cli`` with command line arguments"""
    happi_cli.add_command(config)
    happi_cli()



import inspect
import logging
from collections.abc import Generator
from typing import ClassVar, Optional

import entrypoints

from .item import HappiItem, OphydItem

logger = logging.getLogger(__name__)

HAPPI_ENTRY_POINT_KEY = "happi.containers"


DEFAULT_REGISTRY = {"OphydItem": OphydItem, "HappiItem": HappiItem}


class HappiRegistry:
    """
    Happi Container Registry singleton.

    This registry keeps a mapping of full happi container names to their
    respective :class:`HappiItem` subclasses.

    Entries in the registry are populated in the following order:

    1. From the ``happi.containers.DEFAULT_REGISTRY`` dictionary.  The names
       of the containers are assumed to already include any relevant qualifiers
       such as package names.  That is, the default ``"OphydItem"`` will be
       used as-is and retrieved from the registry by that name alone.
    2. Through Python package-defined entrypoints with the key
       ``happi.containers``.

       For example, consider a package named ``"mypackagename"`` that defines
       the happi container ``ContainerName`` inside of the Python module
       ``mypackagename.happi.containers_module.ContainerName``.
       The fully qualified name - as accessed through the happi client for each
       item in the database - may be customized in the entrypoint.

       Here, we want the container to be accessible by way of
       ``"desired_prefix.ContainerName"``.  The following would be how this
       entrypoint should be configured in ``setup.py``.

    .. code::

        setup(
            name="mypackagename",
            entry_points={
                "happi.containers": [
                    "desired_prefix=mypackagename.happi.containers_module",
                ],
            },
        )

    Containers may also be added manually to the registry by way of:

    .. code::

        import happi

        happi.containers.registry[item_name] = ContainerClass
    """
    #: Has __init__ been called on the registry?
    __initialized: bool
    #: The singleton instance of the HappiRegistry.
    __instance: ClassVar[Optional["HappiRegistry"]] = None
    #: Has the registry been initialized once?
    _loaded: bool
    #: Registry of happi container name to class
    _registry: dict[str, type[HappiItem]]
    #: Registry of happi container class to name
    _reverse_registry: dict[type[HappiItem], str]

    def __init__(self):
        if self.__initialized:
            # This guard ensures that `__init__` is not called twice on the
            # singleton.
            return
        self._registry = {}
        self._reverse_registry = {}
        self._loaded = False
        self.__initialized = True

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = object.__new__(HappiRegistry)
            cls.__instance.__initialized = False
        return cls.__instance

    def __getitem__(self, item: str) -> Optional[type[HappiItem]]:
        if not self._loaded or item not in self._registry:
            self.load()
        return self._registry.get(item)

    def __setitem__(self, item: str, klass: type[HappiItem]) -> None:
        self._safe_add(item, klass)

    def __contains__(self, item: str) -> bool:
        if not self._loaded or item not in self._registry:
            self.load()
        return item in self._registry

    def items(self) -> Generator[tuple[str, type[HappiItem]], None, None]:
        """All (item_name, item_class) entries in the registry."""
        if not self._loaded:
            self.load()
        yield from self._registry.items()

    def entry_for_class(self, klass: type[HappiItem]) -> Optional[str]:
        """
        Get the happi item container name given its class.

        Parameters
        ----------
        klass : HappiItem class
            The class to get the name of.

        Returns
        -------
        str or None
            The full container name, if in the registry.
        """
        if not self._loaded or klass not in self._reverse_registry:
            self.load()
        return self._reverse_registry.get(klass)

    def _safe_add(self, entry_name: str, klass: type[HappiItem]):
        """
        Add and entry into the registry and raise RuntimeError in case a
        duplicated entry is found.

        Parameters
        ----------
        entry_name : str
            The entrypoint identifier
        klass : class
            The class to add

        Raises
        ------
        RuntimeError
            Raises this exception if the entry is duplicated.
        """
        def make_entry_name():
            """
            Cut out the package name and replace with the entrypoint name so it
            can later be moved if needed.

            Parameters
            ----------
            klass : class

            Returns
            -------
            str
                The formatted name.
            """
            module = ".".join(klass.__module__.split(".")[1:])
            return ".".join([entry_name, module, klass.__name__])

        key = make_entry_name()

        class_in_registry = self._registry.get(key, None)
        if class_in_registry is klass:
            return
        if class_in_registry is not None:
            raise RuntimeError(f"Duplicated entry found for key: {key} "
                               f"and class: {klass} {class_in_registry}")
        if klass in self._reverse_registry:
            dup_key = self._reverse_registry.get(klass)
            raise RuntimeError(f"Duplicated entry found. Keys: {key} "
                               f"and {dup_key} point to same class: {klass}")

        self._registry[key] = klass
        self._reverse_registry[klass] = key

    def load(self) -> None:
        """
        Load entries into the Registry.
        """
        def valid_entry(klass):
            # Avoid happi internal classes due to imports
            return inspect.isclass(klass) and issubclass(klass, HappiItem) \
                and not klass.__module__.startswith('happi.')

        for name, klass in DEFAULT_REGISTRY.items():
            if name not in self._registry:
                self._registry[name] = klass
                self._reverse_registry[klass] = name

        _entries = entrypoints.get_group_all(HAPPI_ENTRY_POINT_KEY)

        for entry in _entries:
            try:
                entry_name = entry.name
                obj = entry.load()
            except Exception:
                logger.exception("Failed to load happi.containers entry: %s",
                                 entry)
                continue
            if valid_entry(obj):
                self._safe_add(entry_name, obj)
            elif inspect.ismodule(obj):
                for _, var in inspect.getmembers(obj):
                    if valid_entry(var):
                        self._safe_add(entry_name, var)

        self._loaded = True


registry = HappiRegistry()



import pathlib

from qtpy import QtCore, QtWidgets

import happi
import happi.qt


class HappiDeviceExplorer(QtWidgets.QFrame):
    _GROUP_KEYS = {'Name': 'name',
                   'Type': 'type',
                   'Device Class': 'device_class'}

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.view = happi.qt.model.HappiDeviceTreeView(self)
        self.view.groups = [v for _, v in self._GROUP_KEYS.items()]

        self.group_label = QtWidgets.QLabel("&Group By")
        self.group_combo = QtWidgets.QComboBox()
        self.group_label.setBuddy(self.group_combo)

        for display, val in self._GROUP_KEYS.items():
            self.group_combo.addItem(display, val)

        def set_group(_):
            key = self.group_combo.currentData()
            self.view.group_by(key)

        self.group_combo.currentIndexChanged.connect(set_group)

        self.filter_label = QtWidgets.QLabel("&Filter")
        self.filter_edit = QtWidgets.QLineEdit()
        self.filter_label.setBuddy(self.filter_edit)

        def set_filter(text):
            self.view.proxy_model.setFilterRegExp(text)

        self.filter_edit.textEdited.connect(set_filter)

        self.setLayout(QtWidgets.QVBoxLayout())

        self.filter_frame = QtWidgets.QFrame()
        self.filter_frame.setLayout(QtWidgets.QGridLayout())

        self.filter_frame.layout().addWidget(self.filter_label, 1, 0)
        self.filter_frame.layout().addWidget(self.filter_edit, 1, 1)
        self.filter_frame.layout().addWidget(self.group_label, 2, 0)
        self.filter_frame.layout().addWidget(self.group_combo, 2, 1)

        self.splitter = QtWidgets.QSplitter()
        self.splitter.setOrientation(QtCore.Qt.Vertical)

        self.splitter.addWidget(self.filter_frame)
        self.splitter.addWidget(self.view)
        self.splitter.setSizes([0, 1])
        self.splitter.setHandleWidth(10)
        self.splitter.setStretchFactor(0, 0)
        self.splitter.setStretchFactor(1, 1)

        self.splitter.setCollapsible(1, False)

        handle = self.splitter.handle(1)
        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        button = QtWidgets.QToolButton(handle)
        button.setArrowType(QtCore.Qt.DownArrow)
        button.clicked.connect(lambda: self.handle_splitter_button(True))
        layout.addWidget(button)
        button = QtWidgets.QToolButton(handle)
        button.setArrowType(QtCore.Qt.UpArrow)
        button.clicked.connect(lambda: self.handle_splitter_button(False))
        layout.addWidget(button)
        handle.setLayout(layout)

        self.layout().addWidget(self.splitter)

    def handle_splitter_button(self, down=True):
        if down:
            self.splitter.setSizes([1, 1])
        else:
            self.splitter.setSizes([0, 1])


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    file_path = pathlib.Path(__file__).resolve()
    db_path = file_path.parent.parent / "db.json"

    cli = happi.Client(path=db_path)
    w = HappiDeviceExplorer()
    w.view.client = cli
    w.view.search(type='OphydItem')
    w.show()

    app.exec_()



import ast
import logging
from typing import Optional

import click
import prettytable

from happi.errors import EnforceError, TransferError
from happi.item import EntryInfo
from happi.utils import (OptionalDefault, is_valid_identifier_not_keyword,
                         optional_enforce)

logger = logging.getLogger(__name__)

hopefully_unique_keyword = 'verylonghopefullyuniquekeywordthatdoesnotconflict'


def read_user_dict(prompt, default: Optional[dict] = None):
    """
    Prompt for a dictionary, prompting for keys and values separately
    """
    user_dict = {}
    click.echo(prompt + '\nKey must be a string.  ' +
               'Enter a blank key to complete dict entry.')
    while True:
        key = click.prompt('  key',
                           default=hopefully_unique_keyword,
                           show_default=False,
                           value_proc=is_valid_identifier_not_keyword)
        if key is hopefully_unique_keyword:
            break

        value = click.prompt('  value')
        try:
            value = ast.literal_eval(value)
        except (ValueError, SyntaxError):
            logger.debug(f'Taking {value} as a string')

        user_dict.update({key: value})

    if not user_dict:
        return default

    return user_dict


def enforce_list(value):
    """
    Special handling for list inputs.
    Accept python lists, and attempts to intepret any strings as lists

    Raises
    ------
    EnforceError
        if value cannot be interpreted as a list
    """
    if isinstance(value, list):
        return value

    # if not a list, try to interpret as a list
    try:
        value = ast.literal_eval(value)
    except (ValueError, SyntaxError) as e:
        raise EnforceError(e)

    if not isinstance(value, list):
        raise EnforceError(f'Provided value ({value}) is not a list')
    return value


def prompt_for_entry(entry_info, clone_source=None):
    """Prompt for an entry based on the entry_info provided"""
    if clone_source:
        default = getattr(clone_source, entry_info.key)
    else:
        default = entry_info.default

    if entry_info.optional and (default is None):
        # Prompt will continue to prompt if default is None
        # Provide a dummy value to allow prompt to exit
        default = OptionalDefault()
        enforce_fn = optional_enforce(entry_info.enforce_value)
    else:
        enforce_fn = entry_info.enforce_value

    enforce_str = getattr(entry_info.enforce, '__name__',
                          str(entry_info.enforce))
    val_prompt = (f'Enter value for {entry_info.key}, '
                  f'enforce={enforce_str}')

    # prompt differently depending on the enforce type
    if entry_info.enforce is list:
        logger.debug('prompting for list')
        # special handling for an optional list
        if isinstance(default, OptionalDefault):
            list_enforce = optional_enforce(enforce_list)
        else:
            list_enforce = enforce_list

        value = click.prompt(val_prompt, default=default,
                             value_proc=list_enforce)
    elif entry_info.enforce is dict:
        logger.debug('prompting for dict')
        value = read_user_dict(val_prompt, default=default)
    elif entry_info.enforce is bool:
        logger.debug('prompting for bool')
        # coerces into y or n, preventing random strings (eg. 'f') from
        # evaluating as True
        value = click.confirm(val_prompt, default=default)
    else:
        # everything else is a callable
        value = click.prompt(val_prompt, default=default,
                             value_proc=enforce_fn)

    if isinstance(value, OptionalDefault):
        # Default was None, return None
        value = None

    return value


def transfer_container(client, item, target):
    """
    Interactively step through transferring an item into a new container
    Works by catching exceptions raised by client.change_container
    and prompting for updates.

    Steps:
    - Display information for item and its target container
    - Prepare Entries: prompt to include extra entries and fill missin
                       entries
    - Amend Entries: Attempt to coerce ``item`` into ``target``.  If
                     there is an error, prompt for a fix.
    - Create new item and prompt for confirmation to save.

    Parameters
    ----------
    client : happi.client.Client
        Happi client connected to database and container registry

    item : happi.HappiItem
        Loaded item to transfer

    target : Type[happi.HappiItem]
        Target container for ``item``.  Container constructor
    """
    target_name = target.__name__
    print(f'Attempting to transfer {item.name} to {target_name}...')
    # compare keys in item to target
    item_info = item.post()
    ignore_names = ['type', 'creation', 'last_edit']
    item_public_keys = sorted([key for key in item_info.keys()
                               if not key.startswith('_')])
    matching_keys = [n for n in item_public_keys
                     if n in target.info_names]
    item_exclusive = [n for n in item_public_keys
                      if n not in target.info_names + ignore_names]
    target_exclusive = [n for n in target.info_names
                        if n not in item_public_keys]

    pt = prettytable.PrettyTable()
    pt.field_names = [f'{item.name}', f'{target_name}']
    for n in matching_keys:
        pt.add_row([n, n])
    for ni in item_exclusive:
        pt.add_row([ni, '-'])
    for nt in target_exclusive:
        pt.add_row(['-', nt])

    click.echo(pt)

    edits = {}
    click.echo('\n----------Prepare Entries-----------')
    # Deal with extra keys in item
    for n in item_exclusive:
        extra_prompt = (f'Include entry from {item.name}: '
                        f'{n} = "{item_info.get(n)}"?')
        if click.confirm(extra_prompt, default=True):
            edits.update({n: item_info.get(n)})

    # Deal with missing keys in target, deal with validation later
    for nt in target_exclusive:
        einfo: EntryInfo = getattr(target, nt)
        val = prompt_for_entry(einfo)
        edits.update({nt: val})

    # Actually apply changes and cast item into new container
    # Enforce conditions are dealt with here
    click.echo('\n----------Amend Entries-----------')
    success = False
    new_kwargs = None
    while not success:
        try:
            new_kwargs = client.change_container(item, target, edits=edits)
            success = True
        except TransferError as e:
            print(e)
            einfo = getattr(target, e.key)
            new_val = prompt_for_entry(einfo)
            edits.update({e.key: new_val})

    if not new_kwargs:
        logger.debug('transfer_container failed, no kwargs returned')
        return

    item = client.create_item(target, **new_kwargs)
    item.show_info()

    if click.confirm('Save final item?'):
        logger.debug('deleting original object and replacing')
        client.remove_item(item)
        item.save()



"""
Functions to instantiate the Python representations of happi Containers
"""
import asyncio
import importlib
import logging
import sys
import time
import types
from functools import partial
from multiprocessing.pool import ThreadPool
from typing import Any, Callable, Optional

from jinja2 import Environment, meta

from .item import HappiItem
from .utils import create_alias

logger = logging.getLogger(__name__)

cache = dict()
main_event_loop = None


def fill_template(
    template: str,
    item: HappiItem,
    enforce_type: bool = False
) -> Any:
    """
    Fill a Jinja2 template using information from a happi item.

    Parameters
    ----------
    template : str
        Jinja2 template source.

    item : HappiItem
        Any happi item container.

    enforce_type : bool, optional
        Force the output of the rendered template to match the enforced type of
        the happi information that was used to fill it.

    Returns
    -------
    rendered : Any
        string-rendered template if ``enforce_type`` is False, or if happi is
        unable to cast the rendered value to the given type.
    """
    # Create a template and render our happi information inside it
    env = Environment().from_string(template)
    filled = env.render(**item.post())
    # Find which variable we used in the template, get the type and convert our
    # rendered template to agree with this
    info = meta.find_undeclared_variables(env.environment.parse(template))
    if len(info) != 1 or not enforce_type:
        # Enforcing types only works with 1 attribute name in the template
        return filled

    # Get the original attribute back from the item. If this does not exist
    # there is a possibility it is a piece of metadata e.t.c
    attr_name = info.pop()
    try:
        typed_attr = getattr(item, attr_name)
    except AttributeError:
        logger.warning(
            "Can not enforce type to match attribute %s as it does not "
            "exist on the container.", attr_name
        )
        return filled

    # If this was a straight substitution with nothing else in the template we
    # can just return the attribute itself thus preserving type
    if str(typed_attr) == filled:
        return typed_attr

    # If there is something more complex going on we can attempt to convert
    # it to match the type of the original
    attr_type = type(typed_attr)
    try:
        return attr_type(filled)
    except ValueError:
        logger.exception(
            "Unable to convert attribute %s from %r to %s",
            attr_name, filled, attr_type
        )
    return filled


def from_container(
    item: HappiItem,
    attach_md: bool = True,
    run_post_attach_hook: bool = True,
    use_cache: bool = True,
    threaded: bool = False,
) -> Any:
    """
    Load an object (or "device") from a compatible HappiItem.

    The item container is queried for the ``device_class``, ``args``, and
    ``kwargs``. If the associated Python module is not already loaded it will
    be imported. The specified class is then instantiated with the given args
    and kwargs provided.

    The name ``device`` here refers to what is created after instantiating
    ``device_class``.

    This function does not attempt to catch exceptions either during module
    imports or object creation. If you would like a series of independent
    devices to be loaded use :func:`.load_devices`.

    By default, the instantiated object has the original container added on as
    ``.md``. This allows applications to utilize additional metadata
    information that may not be included in the basic class constructor. On
    later calls, the container you request is checked against this stored
    metadata. If a discrepancy is found the object is **forced** to reload, not
    retrieved from the cache.

    Parameters
    ----------
    item : happi.HappiItem
        The item to load.
    attach_md : bool, optional
        Attach the container to the instantiated object as ``md``.
    run_post_attach_hook : bool, optional
        Run ``post_happi_md`` method on the device after attaching the container
        if possible.
    use_cache : bool, optional
        When devices are loaded they are stored in the ``happi.cache``
        dictionary. This means that repeated attempts to load the device will
        return the same object. This prevents unnecessary EPICS connections
        from being initialized in the same process. If a new object is
        needed, set ``use_cache`` to `False` and a new object will be created,
        overriding the current cached object. An object with matching name
        and differing metadata will always return a new instantiation of the
        device.
    threaded : bool, optional
        Set this to `True` when calling inside a thread.  This is currently
        unused.

    Returns
    -------
    obj : Any
        This will be of the same type as the return value of the provided
        item's ``device_class``.  As that may be a factory function or a
        class, the exact return type is not guaranteed.
    """

    # Return a cached version of the device if present and not forced
    if use_cache and item.name in cache:
        cached_device = cache[item.name]
        # If the metadata has not been modified or we can't review it.
        # Return the cached object
        if hasattr(cached_device, 'md') and cached_device.md == item:
            logger.debug("Loading %s from cache...", item.name)
            return cached_device

        # Otherwise reload
        logger.warning(
            "Device %s has already been loaded, but the database information "
            "has been modified. Reloading...",
            item.name
        )

    # Find the class and module of the container.
    if not item.device_class:
        raise ValueError(
            f"Item {item.name} does not have an associated Python class"
        )

    cls = import_class(item.device_class)

    # Create correctly typed arguments from happi information
    def create_arg(arg):
        if not isinstance(arg, str):
            return arg
        return fill_template(arg, item, enforce_type=True)

    # Treat all our args and kwargs as templates
    args = [create_arg(arg) for arg in item.args]
    kwargs = {key: create_arg(val) for key, val in item.kwargs.items()}
    # maybe filter out null kwargs
    new_kwargs = {}
    kwarg_include = item._info_attrs['kwargs'].include_default_as_kwarg
    for entry, value in kwargs.items():
        einfo = item._info_attrs.get(entry, False)
        # if kwargs are marked as include, let individual entries veto
        if kwarg_include:
            if (not einfo or einfo.include_default_as_kwarg or
                    not einfo.default == value):
                new_kwargs[entry] = value
        # if kwargs are marked to exclude, ignore individual entry setting
        else:
            if (not einfo or not einfo.default == value):
                new_kwargs[entry] = value
    kwargs = new_kwargs

    # Return the instantiated item
    obj = cls(*args, **kwargs)
    # Attach the metadata to the object
    if attach_md:
        try:
            setattr(obj, 'md', item)
        except Exception:
            logger.warning("Unable to attach metadata dictionary to device")

        if run_post_attach_hook and hasattr(obj, 'post_happi_md'):
            try:
                obj.post_happi_md()
            except Exception as ex:
                logger.warning(f"Unable to run {obj}.post_happi_md: {ex}")

    # Store the device in the cache
    cache[item.name] = obj
    return obj


def import_class(device_class: str):
    """
    Interpret a device class import string and extract the class object.

    Parameters
    ----------
    device_class : str
        The module path to find the class e.g.
        ``"pcdsdevices.device_types.IPM"``.

    Returns
    -------
    cls : type or callable
        The class or factory function referred to by the input string.
    """

    mod, cls = device_class.rsplit('.', 1)
    # Import the module if not already present
    # Otherwise use the stashed version in sys.modules
    if mod in sys.modules:
        logger.debug("Using previously imported version of %s", mod)
        mod = sys.modules[mod]
    else:
        logger.debug("Importing %s", mod)
        mod = importlib.import_module(mod)
    # Gather our device class from the given module
    try:
        return getattr(mod, cls)
    except AttributeError as exc:
        raise ImportError("Unable to import %s from %s" %
                          (cls, mod.__name__)) from exc


PostLoad = Callable[[Any], None]


def load_devices(
    *items: HappiItem,
    pprint: bool = False,
    namespace: Optional[object] = None,
    use_cache: bool = True,
    threaded: bool = False,
    post_load: Optional[PostLoad] = None,
    include_load_time: bool = False,
    load_time_threshold: float = 0.5,
    **kwargs
):
    """
    Load a series of devices by way of their HappiItems into a namespace.

    Parameters
    ----------
    *items : HappiItem
        List of happi containers to load.
    pprint: bool, optional
        Print results of device loads.
    namespace : object, optional
        Namespace to collect loaded devices in. By default this will be a
        `types.SimpleNamespace`.
    use_cache : bool, optional
        If set to `False`, we'll ignore the cache and always make new
        devices.
    threaded : bool, optional
        Set to `True` to create each device in a background thread.  Note that
        this assumes that no two devices provided are the same. You are not
        guaranteed to load from the cache correctly if you ask for the same
        device to be loaded twice in the same threaded load.
    post_load : function, optional
        Function of one argument to run on each device after instantiation.
        This could be especially useful during the threaded loading process.
    include_load_time : bool, optional
        Include load time in each message.
    load_time_threshold : float, optional
        Load time above this value, in seconds, will be shown if
        ``include_load_time`` is set.
    kwargs
        Additional keyword arguments passed to :func:`.from_container`.
    """

    # Create our namespace if we were not given one
    namespace = namespace or types.SimpleNamespace()
    name_list = [container.name for container in items]
    if threaded:
        # Pre-import because imports in threads have race conditions
        for item in items:
            try:
                import_class(item.device_class)
            except Exception:
                # Just wait for the normal error handling later
                pass
        global main_event_loop
        if main_event_loop is None:
            main_event_loop = asyncio.get_event_loop()
        pool = ThreadPool(len(items))
        opt_load = partial(
            load_device,
            pprint=pprint,
            use_cache=use_cache,
            threaded=True,
            post_load=post_load,
            include_load_time=include_load_time,
            load_time_threshold=load_time_threshold,
            **kwargs
        )
        loaded_list = pool.map(opt_load, items)
    else:
        loaded_list = []
        for item in items:
            loaded = load_device(
                item,
                pprint=pprint,
                use_cache=use_cache,
                threaded=False,
                post_load=post_load,
                include_load_time=include_load_time,
                load_time_threshold=load_time_threshold,
                **kwargs
            )
            loaded_list.append(loaded)
    for dev, name in zip(loaded_list, name_list):
        attr = create_alias(name)
        setattr(namespace, attr, dev)
    return namespace


def load_device(
    item: HappiItem,
    pprint: bool = False,
    threaded: bool = False,
    post_load: Optional[PostLoad] = None,
    include_load_time: bool = False,
    load_time_threshold: float = 0.5,
    **kwargs
) -> Any:
    """
    Call :func:`.from_container ` and show success/fail.

    Parameters
    ----------
    item : happi.HappiItem
        HappiItem to be loaded.
    pprint: bool, optional
        Print results of the loading status.
    threaded: bool, optional
        Set this to `True` when calling inside a thread.
    post_load : function, optional
        Function of one argument to run on each object after instantiation.
        This could be especially useful during the threaded loading process.
    include_load_time : bool, optional
        Include load time in each message.
    load_time_threshold : float, optional
        Load time above this value, in seconds, will be shown if
        ``include_load_time`` is set.
    kwargs:
        Additional keyword arguments passed to :func:`.from_container`.

    Returns
    -------
    obj : Any
        Returns either the loaded object or the resulting exception from
        trying to load the object.
    """

    logger.debug("Loading device from item %s ...", item.name)

    # We sync with the main thread's loop so that they work as expected later
    if threaded and main_event_loop is not None:
        asyncio.set_event_loop(main_event_loop)

    load_message = f"Loading {item.name} [{item.device_class}] ... "
    success = "\033[32mSUCCESS\033[0m!"
    failed = "\033[31mFAILED\033[0m"

    start_time = time.monotonic()
    if pprint and not threaded:
        print(load_message, end='')

    def get_load_time() -> str:
        """Get the load time information for display."""
        elapsed_time = time.monotonic() - start_time
        if include_load_time and elapsed_time >= load_time_threshold:
            return f" ({elapsed_time:.1f} s)"
        return ""

    def print_load_message(message: str, elapsed: str) -> None:
        if not pprint:
            return

        if threaded:
            print(f"{load_message} {message}{elapsed}")
        else:
            print(f"{message}{elapsed}")

    try:
        obj = from_container(item, **kwargs)
        if post_load is not None:
            post_load(obj)
    except Exception as exc:
        elapsed = get_load_time()
        print_load_message(failed, elapsed)
        logger.exception("Error loading %s%s", item.name, elapsed)
        return exc

    elapsed = get_load_time()
    logger.info("%s %s%s", load_message, success, elapsed)
    print_load_message(success, elapsed)
    return obj



import argparse
import logging
import os
import subprocess
from distutils.spawn import find_executable

from jinja2 import Environment

import happi

logger = logging.getLogger(__name__)

description = """\
Launch the EDL screen for a Happi device based on the information contained
within the database

Example:
    %(prog)s XRT_M2  --embedded

This launches the embedded EDL screen for the device contained within the Happi
database with the name 'XRT_M2'
"""


def launch(path, wait=True, wd=None, macros=None):
    """
    Launch an EDL file

    Parameters
    ----------
    path : str
        Path to file

    wd : str, optional
        Working directory to launch screen, otherwise the current directory
        is used

    wait : bool, optional
        Block the main thread while the EDM preview is open

    macros : str, optional
        String of macro substitutions

    Returns
    -------
    proc : ``subprocess.Popen``
        Process containing EDM launch

    Raises
    ------
    FileNotFoundError:
        If the .edl file does not exist

    OSError:
        If the ``edm`` executable is not in the system path

     Example
    -------
    .. code::

        edm_proc = launch('path/to/my.edl', MACRO='TST:MACRO')
    """
    if not os.path.exists(path):
        raise FileNotFoundError(path)

    edm_args = ['edm', '-x', '-eolc']

    if macros:
        edm_args.extend(['-m', macros])

    edm_args.append(path)

    try:
        print("Launching {} with the following arguments {} ..."
              "".format(path, edm_args))
        proc = subprocess.Popen(edm_args, cwd=wd, stdout=None)

        if wait:
            proc.wait()

    except OSError:
        if not find_executable('edm'):
            raise OSError('EDM is not in current environment')

        raise

    except KeyboardInterrupt:
        print('Aborted via KeyboardInterrupt ...')
        proc.terminate()

    return proc


def main():
    # Parse command line
    fclass = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(epilog=description,
                                     formatter_class=fclass)

    # Arguments
    parser.add_argument('name', help='Name of Device')

    parser.add_argument('-e', '--embedded', dest='embedded',
                        help='Choice to use embedded screen',
                        default=False, action='store_true')

    parser.add_argument('-b', '--block', dest='block',
                        help='Block the main thread while '
                             'the screen is open',
                        default=False, action='store_true')

    parser.add_argument('-d', '--dir', dest='dir',
                        help='Directory to launch screen from',
                        default=None)

    # Parse arguments
    args = parser.parse_args()
    client = happi.Client()

    try:
        device = client.load_device(name=args.name)

    except happi.errors.SearchError:
        print('Unable to locate any device with '
              'name {} in the database.'
              ''.format(args.name))
        return

    # Create string of macros based on template
    env = Environment().from_string(device.macros)
    macros = env.render(**device.post())

    # Gather screen path
    if args.embedded:
        screen = device.embedded_screen
    else:
        screen = device.main_screen

    # Launch screen
    launch(screen, wait=args.block, wd=args.dir, macros=macros)


if __name__ == '__main__':
    main()



# test_cli.py

import functools
import itertools
import logging
import re
from collections.abc import Iterable
from typing import Any
from unittest import mock

import click
import IPython
import pytest
from click.testing import CliRunner

import happi
from happi.cli import happi_cli, search
from happi.errors import SearchError

try:
    import pcdsutils.profile
    if pcdsutils.profile.has_line_profiler:
        test_line_prof = True
    else:
        test_line_prof = False
except ImportError:
    test_line_prof = False


logger = logging.getLogger(__name__)


@pytest.fixture(scope='function')
def client(happi_cfg: str):
    return happi.client.Client.from_config(cfg=happi_cfg)


def trim_split_output(strings: str, delim: str = '\n'):
    """
    Trim output of timestamped messages and registry lists.
    Split on delimiter (newline by default)
    """
    date_pattern = r"\[(\d{4})[-](0[1-9]|1[012])[-].*\]"

    omit_substrs = [r"happi.containers.", r"- DEBUG -"]

    # remove registry items
    new_out = strings.split(delim)

    # strip registry items from outside entrypoints?
    new_out = [st for st in new_out
               if not any([re.search(substr, st) for substr in omit_substrs])]
    # strip date-time from logging messages
    new_out = [re.sub(date_pattern, '', st) for st in new_out]
    return new_out


def assert_match_expected(
    result: click.testing.Result,
    expected_output: Iterable[str],
    expected_return: int = 0,
    raise_failures: bool = False
):
    """
    standard checks for a cli result, confirms output matches expected
    SystemExit exception is acceptable if the expected return code is some failure code
    """
    logger.debug(result.output)
    assert result.exit_code == expected_return
    assert not result.exception or (isinstance(result.exception, SystemExit) and
                                    expected_return != 0)

    trimmed_output = trim_split_output(result.output)
    for message, expected in zip(trimmed_output, expected_output):
        if raise_failures:
            assert message == expected
        else:
            if not message == expected:
                print('Output does not match expected value:\n'
                      f'- Output: {message}\n- Expected: {expected}')


def assert_in_expected(
    result: click.testing.Result,
    expected_inclusion: str,
    expected_return: int = 0
):
    """
    standard checks for a cli result, confirms output matches expected
    SystemExit exception is acceptable if the expected return code is some failure code
    """
    logger.debug(result.output)
    assert result.exit_code == expected_return
    assert not result.exception or (isinstance(
        result.exception, SystemExit) and expected_return != 0)
    assert expected_inclusion in result.output


def test_cli_no_argument(runner: CliRunner):
    result = runner.invoke(happi_cli)
    assert result.exit_code == 0
    assert result.exception is None
    assert 'Usage:' in result.output
    assert 'Options:' in result.output
    assert 'Commands:' in result.output


def test_repair_name(runner: CliRunner, bad_happi_cfg: str):
    search_string = '_id=tst_id'

    # make sure the name and id are different
    with search.make_context('search', [search_string], obj=bad_happi_cfg) as ctx:
        res = search.invoke(ctx)
    assert all([r['_id'] != r['name'] for r in res])

    # run repair
    runner.invoke(happi_cli, ['--path', bad_happi_cfg, 'repair', search_string])

    # make sure the name and id are the same
    with search.make_context('search', [search_string], obj=bad_happi_cfg) as ctx:
        res = search.invoke(ctx)
    assert all([r['_id'] == r['name'] for r in res])


def test_repair_no_save_if_no_changes(runner: CliRunner, happi_cfg: str):
    # some arbitrary entry
    search_string = '_id=tst_base_pim'

    # get last edit time of item, so can make sure it isn't changed by repair
    with search.make_context('search', [search_string], obj=happi_cfg) as ctx:
        res = search.invoke(ctx)
    pre_repair_edit_time = res[0]['last_edit']

    # run repair, which should find it doesn't need to repair the item's name and so doesn't need to save
    runner.invoke(happi_cli, ['--path', happi_cfg, 'repair', search_string])

    with search.make_context('search', [search_string], obj=happi_cfg) as ctx:
        res = search.invoke(ctx)
    after_repair_edit_time = res[0]['last_edit']

    # should not have saved, so edit time should be unchanged
    assert after_repair_edit_time == pre_repair_edit_time


def test_search(client: happi.client.Client, happi_cfg: str):
    res = client.search_regex(beamline="TST")

    with search.make_context('search', ['beamline=TST'], obj=happi_cfg) as ctx:
        res_cli = search.invoke(ctx)

    assert [r.item for r in res] == [r.item for r in res_cli]


def test_search_with_name(
    client: happi.client.Client,
    runner: CliRunner,
    happi_cfg: str
):
    res = client.search_regex(name='TST_BASE_PIM2')

    with search.make_context('search', ['TST_BASE_PIM2'], obj=happi_cfg) as ctx:
        res_cli = search.invoke(ctx)

    assert [r.item for r in res] == [r.item for r in res_cli]
    # test duplicate search parameters
    bad_result = runner.invoke(happi_cli, ['--path', happi_cfg,
                                           'search', 'TST_BASE_PIM2',
                                           'name=TST_BASE_PIM2'])

    assert bad_result.exit_code == 1
    assert 'duplicate search criteria' in bad_result.output


def test_search_glob_regex(runner: CliRunner, happi_cfg: str):
    glob_result = runner.invoke(happi_cli, ['--path', happi_cfg, 'search',
                                '--names', 'tst_*2'])

    regex_result = runner.invoke(happi_cli, ['--path', happi_cfg, 'search',
                                 '--names', '--regex', r'tst_.*\d'])

    assert glob_result.output == regex_result.output


def test_search_z(happi_cfg: str, client: happi.client.Client):
    res = client.search_regex(z="6.0")
    with search.make_context('search', ['z=6.0'], obj=happi_cfg) as ctx:
        res_cli = search.invoke(ctx)

    assert [r.item for r in res] == [r.item for r in res_cli]


def test_search_z_range(
    client: happi.client.Client,
    runner: CliRunner,
    happi_cfg: str
):
    res = client.search_range('z', 3.0, 6.0)

    with search.make_context('search', ['z=3.0,6.0'], obj=happi_cfg) as ctx:
        res_cli = search.invoke(ctx)

    assert [r.item for r in res] == [r.item for r in res_cli]

    # test range intersection
    result = runner.invoke(happi_cli, ['--path', happi_cfg, 'search',
                           '--names', 'z=3.0,6.1', 'y=9.0,12.0'])

    assert result.exit_code == 0
    assert result.output == 'tst_base_pim2\n'

    # test invalid range
    bad_result = runner.invoke(happi_cli, ['--path', happi_cfg,
                               'search', 'z=6.0,3.0'])

    assert bad_result.exit_code == 1
    assert "Invalid range, make sure start < stop" in bad_result.output

    # test conflicting ranges (should return no items)
    conflict_result = runner.invoke(happi_cli, ['--path', happi_cfg,
                                    'search', 'y=1,3', 'z=3.0,6.0'])

    assert conflict_result.exit_code == 0
    assert 'No items found' in conflict_result.output

    # test conflicting ranges but with opposite order
    conflict_result = runner.invoke(happi_cli, ['--path', happi_cfg,
                                    'search', 'z=3.0,6.0', 'y=1,3'])

    assert conflict_result.exit_code == 0
    assert 'No items found' in conflict_result.output


def test_search_int_float(runner: CliRunner, happi_cfg: str):
    int_result = runner.invoke(happi_cli, ['--path', happi_cfg,
                               'search', '--names', 'z=3'])
    float_result = runner.invoke(happi_cli, ['--path', happi_cfg,
                                 'search', '--names', 'z=3.0'])
    assert int_result.output == float_result.output

    # # TODO: add this test case once edit works on extraneous info
    # edit_result = runner.invoke(
    #     happi_cli,
    #     ['-v', '--path', happi_cfg, 'edit', 'tst_base_pim', 'z=3.001'],
    #     input='y'
    # )
    # assert edit_result.exit_code == 0

    # int_result = runner.invoke(happi_cli, ['--path', happi_cfg,
    #                            'search', '--names', 'z=3'])
    # float_result = runner.invoke(happi_cli, ['--path', happi_cfg,
    #                              'search', '--names', 'z=3.0'])
    # assert int_result.output == ''
    # assert float_result.output == ''


def test_both_range_and_regex_search(happi_cfg: str, client: happi.client.Client):
    # we're only interested in getting this entry (TST_BASE_PIM2)
    res = client.search_regex(z='6.0')
    # we're going to search for z=3,7 name=TST_BASE_PIM2
    # we should only get in return one entry, even though the z value found 2
    with search.make_context('search', ['name=TST_BASE_PIM2', 'z=3.0,7.0'],
                             obj=happi_cfg) as ctx:
        res_cli = search.invoke(ctx)

    assert [r.item for r in res] == [r.item for r in res_cli]


def test_search_json(runner: CliRunner, happi_cfg: str):
    expected_output = '''[
  {
    "name": "tst_base_pim2",
    "device_class": "types.SimpleNamespace",
    "args": [],
    "kwargs": {
      "name": "{{name}}"
    },
    "active": true,
    "documentation": null,
    "_id": "tst_base_pim2",
    "beamline": "TST",
    "creation": "Wed Jan 30 09:46:00 2019",
    "last_edit": "Fri Apr 13 14:40:08 2018",
    "macros": null,
    "parent": null,
    "prefix": "TST:BASE:PIM2",
    "screen": null,
    "stand": "BAS",
    "system": "diagnostic",
    "type": "HappiItem",
    "z": 6.0,
    "y": 10.0
  }
]'''

    result = runner.invoke(
        happi_cli, ['--path', happi_cfg, 'search', '-j', 'TST_BASE_PIM2']
    )

    assert result.exit_code == 0
    assert_match_expected(result, expected_output.split('\n'), raise_failures=True)


@pytest.mark.parametrize("from_user, expected_output, expected_return", [
    # Test add item - succeeding
    pytest.param('\n'.join(['HappiItem', 'happi_nname', 'device_class',
                            '', '', 'Y', 'docs', 'y']), (
        'Please select a container, or press enter for generic '
        'Ophyd Device container: ', 'OphydItem', 'HappiItem', '',
        'Selection [OphydItem]: HappiItem',
        'Enter value for name, enforce=is_valid_identifier_not_keyword: '
        'happi_nname',
        'Selecting value: happi_nname',
        'Enter value for device_class, enforce=str [optional]: '
        'device_class',
        'Selecting value: device_class',
        'Enter value for args, enforce=list [[]]: ',
        'Selecting value: []',
        'Enter value for kwargs, enforce=dict',
        'Key must be a string.  Enter a blank key to complete dict entry.',
        '  key: ', 'Selecting value: {}',
        'Enter value for active, enforce=bool [Y/n]: Y',
        'Selecting value: True',
        'Enter value for documentation, enforce=str [optional]: docs',
        'Selecting value: docs',
        'Please confirm the item info is correct [y/N]: y',
        ' - INFO -  Adding item',
        ' - INFO -  Storing item HappiItem (name=happi_nname) ...',
        ' - INFO -  Adding / Modifying information for happi_nname ...',
        ' - INFO -  HappiItem HappiItem (name=happi_nname) has been '
        'succesfully added to the database'),
        0
    ),
    # Test add item - aborting
    pytest.param('\n'.join(['HappiItem', 'happi_name2', 'device_class',
                            '["arg1", "arg2"]', 'name', 'my_name', '',
                            'Y', 'docs', 'N']), (
        'Please select a container, or press enter for generic '
        'Ophyd Device container: ', 'OphydItem', 'HappiItem', '',
        'Selection [OphydItem]: HappiItem',
        'Enter value for name, enforce=is_valid_identifier_not_keyword: '
        'happi_name2',
        'Selecting value: happi_name2',
        'Enter value for device_class, enforce=str [optional]: '
        'device_class',
        'Selecting value: device_class',
        'Enter value for args, enforce=list [[]]: ["arg1", "arg2"]',
        "Selecting value: ['arg1', 'arg2']",
        'Enter value for kwargs, enforce=dict',
        'Key must be a string.  Enter a blank key to complete dict entry.',
        '  key: name', '  value: my_name', '  key: ',
        "Selecting value: {'name': 'my_name'}",
        'Enter value for active, enforce=bool [Y/n]: Y',
        'Selecting value: True',
        'Enter value for documentation, enforce=str [optional]: docs',
        'Selecting value: docs',
        'Please confirm the item info is correct [y/N]: N',
        ' - INFO -  Aborting'),
        0
    ),
    # Test add item - invalid container
    pytest.param('HappiInvalidItem', (
        'Please select a container, or press enter for generic '
        'Ophyd Device container: ', 'OphydItem', 'HappiItem', '',
        'Selection [OphydItem]: HappiInvalidItem',
        'Error: Invalid item container HappiInvalidItem'),
        1
    ),
    # Test add item - no reponse, not an optional field,
    # invalid value, add OphydItem
    pytest.param('\n'.join(['', '7', 'ophyd_name', 'device_class',
                            "['arg1', 'arg2']", 'name', 'my_name', '',
                            'Y', 'docs', '', 'some_prefix', 'y']), (
        'Please select a container, or press enter for generic '
        'Ophyd Device container: ', 'OphydItem', 'HappiItem', '',
        'Selection [OphydItem]: ',
        'Enter value for name, enforce=is_valid_identifier_not_keyword: 7',
        'Error: 7 is either not a valid Python identifier, or is a '
        'reserved keyword.',
        'Enter value for name, enforce=is_valid_identifier_not_keyword: '
        'ophyd_name',
        'Selecting value: ophyd_name',
        'Enter value for device_class, enforce=str [optional]: device_class',
        'Selecting value: device_class',
        "Enter value for args, enforce=list [['{{prefix}}']]: "
        "['arg1', 'arg2']", "Selecting value: ['arg1', 'arg2']",
        'Enter value for kwargs, enforce=dict',
        'Key must be a string.  Enter a blank key to complete dict entry.',
        '  key: name', '  value: my_name', '  key: ',
        "Selecting value: {'name': 'my_name'}",
        'Enter value for active, enforce=bool [Y/n]: Y',
        'Selecting value: True',
        'Enter value for documentation, enforce=str [optional]: docs',
        'Selecting value: docs',
        'Enter value for prefix, enforce=str: ',
        'Enter value for prefix, enforce=str: some_prefix',
        'Selecting value: some_prefix',
        'Please confirm the item info is correct [y/N]: y',
        ' - INFO -  Adding item',
        ' - INFO -  Storing item OphydItem (name=ophyd_name) ...',
        ' - INFO -  Adding / Modifying information for ophyd_name ...',
        ' - INFO -  HappiItem OphydItem (name=ophyd_name) has been '
        'succesfully added to the database'),
        0
    )
], ids=["add_succeeding", "add_aborting",
        "add_invalid_container", "add_not_optional_field"])
def test_add_cli(
    from_user: str,
    expected_output: tuple[str, ...],
    expected_return: int,
    happi_cfg: str,
    runner: CliRunner
):
    result = runner.invoke(happi_cli, ['--path', happi_cfg, 'add'],
                           input=from_user)
    assert_match_expected(result, expected_output, expected_return)
    # check this has been added
    dev_name = from_user[1]
    if '_' not in dev_name:  # special case for errored
        return
    client = happi.client.Client.from_config(cfg=happi_cfg)
    item = client.find_item(name=dev_name)
    assert item["_id"]


def test_delete_cli(
    happi_cfg: str,
    runner: CliRunner
):
    delete_result = runner.invoke(
        happi_cli,
        ['--path', happi_cfg, 'delete', 'tst_base_pim'],
        input='y\n'
    )
    assert delete_result.exit_code == 0

    # confirm item is gone
    search_result = runner.invoke(
        happi_cli, ['--path', happi_cfg, 'search', 'tst_base_pim']
    )
    assert_in_expected(search_result, 'No items found', 0)


@pytest.mark.parametrize("from_user, expected_output", [
    # Test add --clone item - succeeding
    pytest.param('\n'.join(['happi_new_name', 'device_class',
                            "['arg1', 'arg2']", 'name', 'my_name', '',
                            '', '', 'y']), [
        'Enter value for name, enforce=is_valid_identifier_not_keyword '
        '[happi_name]: happi_new_name', 'Selecting value: happi_new_name',
        'Enter value for device_class, enforce=str [device_class]: '
        'device_class', 'Selecting value: device_class',
        "Enter value for args, enforce=list [['arg1', 'arg2']]: "
        "['arg1', 'arg2']", "Selecting value: ['arg1', 'arg2']",
        'Enter value for kwargs, enforce=dict',
        'Key must be a string.  Enter a blank key to complete dict entry.',
        '  key: name', '  value: my_name', '  key: ',
        "Selecting value: {'name': 'my_name'}", 'Enter value for active, '
        'enforce=bool [Y/n]: ', 'Selecting value: True',
        'Enter value for documentation, enforce=str [docs]: ',
        'Selecting value: docs', 'Please confirm the item info is '
        'correct [y/N]: y', ' - INFO -  Adding item',
        ' - INFO -  Storing item HappiItem (name=happi_new_name) ...',
        ' - INFO -  Adding / Modifying information for happi_new_name ...',
        ' - INFO -  HappiItem HappiItem (name=happi_new_name) has been '
        'succesfully added to the database', ''], id="clone_succeeding",
    )])
def test_add_clone(
    from_user: str,
    expected_output: tuple[str, ...],
    happi_cfg: str,
    runner: CliRunner
):
    item_info = "\n".join(
        [
            "HappiItem",
            "happi_name",
            "device_class",
            "['arg1', 'arg2']",
            "name",
            "my_name",
            "",
            "Y",
            "docs",
            "y",
        ]
    )
    # add item first
    add_result = runner.invoke(happi_cli, ['--path', happi_cfg, 'add'],
                               input=item_info)
    assert add_result.exit_code == 0

    clone_result = runner.invoke(
        happi_cli, ['--path', happi_cfg, 'add', '--clone', 'happi_name'],
        input=from_user)

    assert_match_expected(clone_result, expected_output)
    client = happi.client.Client.from_config(cfg=happi_cfg)
    item = client.find_item(name="happi_new_name")
    assert item["type"] == 'HappiItem'
    assert item["device_class"] == 'device_class'


def test_add_clone_item_not_found(happi_cfg: str, runner: CliRunner):
    result = runner.invoke(
        happi_cli,
        ['--verbose', '--path', happi_cfg, 'add', '--clone', 'happi_name']
    )
    assert isinstance(result.exception, SearchError)


@pytest.mark.parametrize("from_user, fields, values", [
    # simple edit
    (["documentation=new_docs"], ['documentation'], ['new_docs']),
    # edit kwargs with dictionary
    (["kwargs={\'name\':\'fart\'}"], ['kwargs'], [{'name': 'fart'}]),
    # edit argument list
    (['args=[1,2,3,4]'], ['args'], [[1, 2, 3, 4]]),
    # change multiple entries
    (["active=False", "documentation=yes"], ['active', 'documentation'],
     [False, 'yes'])
])
def test_edit(
    from_user: list[str],
    fields: list[str],
    values: list[Any],
    client: happi.client.Client,
    happi_cfg: str,
    runner: CliRunner
):
    item_info = "\n".join(
        [
            "HappiItem",
            "happi_name",
            "device_class",
            "['arg1', 'arg2']",
            "name",
            "my_name",
            "",
            "Y",
            "docs",
            "y",
        ]
    )
    # add item first
    add_result = runner.invoke(
        happi_cli,
        ['--verbose', '--path', happi_cfg, 'add'],
        input=item_info
    )
    assert add_result.exit_code == 0

    # try edit the previous added item
    edit_result = runner.invoke(
        happi_cli,
        ['--path', happi_cfg, 'edit', 'happi_name', *from_user],
        input='y'  # to confirm edit
    )
    assert edit_result.exit_code == 0

    res = client.search_regex(name='happi_name')
    assert len(res) > 0
    # check fields and values have been modifiedj
    # verify fields match expected python types
    for new_field, new_value in zip(fields, values):
        assert getattr(res[0].item, new_field) == new_value


@pytest.mark.parametrize("edit_args", [
    ["some_invalid_field=sif"],  # invalid field
    ["name=2"],  # bad value for name
    ["kwargs={\'str\':\'beh\'}"],  # bad key in kwarg
])
def test_bad_edit(edit_args: list[str], happi_cfg: str, runner: CliRunner):
    # Test invalid field, note the name is changed to new_name
    bad_edit_result = runner.invoke(
        happi_cli,
        ['--path', happi_cfg, 'edit', 'TST_BASE_PIM', *edit_args]
    )
    assert bad_edit_result.exit_code == 1


def test_load_one_arg(
    caplog: pytest.LogCaptureFixture,
    client: happi.client.Client,
    happi_cfg: str,
    runner: CliRunner
):
    item_info = "\n".join(
        [
            "HappiItem",
            "happi_name",
            "types.SimpleNamespace",
            "",
            "name",
            "my_name",
            "",
            "y",
            "docs",
            "y",
        ]
    )
    # add item first
    add_result = runner.invoke(happi_cli, ['--path', happi_cfg, 'add'],
                               input=item_info)
    assert add_result.exit_code == 0

    # try to load the item
    devices = {}
    devices['happi_name'] = client.load_device(name='happi_name')
    with mock.patch.object(IPython, 'start_ipython') as m:
        _ = runner.invoke(
            happi_cli, ['--path', happi_cfg, 'load', 'happi_name']
        )
        m.assert_called_once_with(argv=['--quick'], user_ns=devices)
    with caplog.at_level(logging.INFO):
        assert "Creating shell with devices" in caplog.text


def test_load_multiple_args(
    caplog: pytest.LogCaptureFixture,
    client: happi.client.Client,
    happi_cfg: str,
    runner: CliRunner
):
    # try to load the item
    devices = {}
    devices['tst_base_pim'] = client.load_device(name='tst_base_pim')
    devices['tst_base_pim2'] = client.load_device(name='tst_base_pim2')
    with mock.patch.object(IPython, 'start_ipython') as m:
        _ = runner.invoke(
            happi_cli, ['--path', happi_cfg, 'load', 'tst_base_pim', 'tst_base_pim2']
        )
        m.assert_called_once_with(argv=['--quick'], user_ns=devices)
    with caplog.at_level(logging.INFO):
        assert "Creating shell with devices" in caplog.text


def test_load_glob_one_arg(
    caplog: pytest.LogCaptureFixture,
    client: happi.client.Client,
    happi_cfg: str,
    runner: CliRunner
):
    # try to load the item
    devices = {}
    devices['tst_base_pim'] = client.load_device(name='tst_base_pim')
    devices['tst_base_pim2'] = client.load_device(name='tst_base_pim2')
    devices['tst_minimal'] = client.load_device(name='tst_minimal')

    with mock.patch.object(IPython, 'start_ipython') as m:
        _ = runner.invoke(
            happi_cli, ['--path', happi_cfg, 'load', 'tst_*']
        )
        m.assert_called_once_with(argv=['--quick'], user_ns=devices)

    with caplog.at_level(logging.INFO):
        assert "Creating shell with devices" in caplog.text


def test_load_glob_multiple_args(
    caplog: pytest.LogCaptureFixture,
    client: happi.client.Client,
    happi_cfg: str,
    runner: CliRunner
):
    # try to load the item
    devices = {}
    devices['tst_base_pim'] = client.load_device(name='tst_base_pim')
    devices['tst_base_pim2'] = client.load_device(name='tst_base_pim2')
    devices['tst_minimal'] = client.load_device(name='tst_minimal')

    with mock.patch.object(IPython, 'start_ipython') as m:
        _ = runner.invoke(
            happi_cli, ['--path', happi_cfg, 'load', 'tst_*', 'device_class=types*']
        )
        m.assert_called_once_with(argv=['--quick'], user_ns=devices)

    with caplog.at_level(logging.INFO):
        assert "Creating shell with devices" in caplog.text


@pytest.mark.parametrize("wrapper", ['{{ "tst_base_pim2":{} }}', "[{}]"])
def test_update(wrapper: str, happi_cfg: str, runner: CliRunner):
    new = """{
        "_id": "tst_base_pim2",
        "active": true,
        "args": [
            "{{prefix}}"
        ],
        "beamline": "TST",
        "creation": "Wed Jan 30 09:46:00 2019",
        "device_class": "types.SimpleNamespace",
        "kwargs": {
            "name": "{{name}}"
        },
        "last_edit": "Fri Apr 13 14:40:08 2018",
        "macros": null,
        "name": "tst_base_pim2",
        "parent": null,
        "prefix": "TST:BASE:PIM2",
        "screen": null,
        "stand": "BAS",
        "system": "diagnostic",
        "type": "OphydItem",
        "z": 6.0,
        "from_update": true
    }"""

    payload = wrapper.format(new)

    result = runner.invoke(happi_cli, ["--path", happi_cfg, "update", payload])
    assert result.exit_code == 0
    client = happi.client.Client.from_config(cfg=happi_cfg)
    item = client.find_item(name="tst_base_pim2")
    assert item["from_update"]


@pytest.mark.parametrize("wrapper", ['{{ "tst_base_pim2":{} }}', "[{}]"])
def test_update_stdin(wrapper: str, happi_cfg: str, runner: CliRunner):
    new = """{
        "_id": "tst_base_pim2",
        "active": true,
        "args": [
            "{{prefix}}"
        ],
        "beamline": "TST",
        "creation": "Wed Jan 30 09:46:00 2019",
        "device_class": "types.SimpleNamespace",
        "kwargs": {
            "name": "{{name}}"
        },
        "last_edit": "Fri Apr 13 14:40:08 2018",
        "macros": null,
        "name": "tst_base_pim2",
        "parent": null,
        "prefix": "TST:BASE:PIM2",
        "screen": null,
        "stand": "BAS",
        "system": "diagnostic",
        "type": "OphydItem",
        "z": 6.0,
        "from_update": true
    }"""

    payload = wrapper.format(new)

    result = runner.invoke(happi_cli, ["--path", happi_cfg, "update", "-"],
                           input=payload)
    assert result.exit_code == 0
    client = happi.client.Client.from_config(cfg=happi_cfg)
    item = client.find_item(name="tst_base_pim2")
    assert item["from_update"]


@pytest.mark.parametrize("from_user, expected_output", [
    pytest.param('\n'.join(['n', 'n', 'n', 'n', 'N', 'n', 'n', 'n', 'y']), [
        'Attempting to transfer tst_base_pim to OphydItem...',
        '+---------------+---------------+',
        '|  tst_base_pim |   OphydItem   |',
        '+---------------+---------------+',
        '|     active    |     active    |',
        '|      args     |      args     |',
        '|  device_class |  device_class |',
        '| documentation | documentation |',
        '|     kwargs    |     kwargs    |',
        '|      name     |      name     |',
        '|     prefix    |     prefix    |',
        '|    beamline   |       -       |',
        '|     macros    |       -       |',
        '|     parent    |       -       |',
        '|     screen    |       -       |',
        '|     stand     |       -       |',
        '|     system    |       -       |',
        '|       y       |       -       |',
        '|       z       |       -       |',
        '+---------------+---------------+', '',
        '----------Prepare Entries-----------',
        'Include entry from tst_base_pim: beamline = "TST"? [Y/n]: n',
        'Include entry from tst_base_pim: macros = "None"? [Y/n]: n',
        'Include entry from tst_base_pim: parent = "None"? [Y/n]: n',
        'Include entry from tst_base_pim: screen = "None"? [Y/n]: n',
        'Include entry from tst_base_pim: stand = "BAS"? [Y/n]: N',
        'Include entry from tst_base_pim: system = "diagnostic"? [Y/n]: n',
        'Include entry from tst_base_pim: y = "40.0"? [Y/n]: n',
        'Include entry from tst_base_pim: z = "3.0"? [Y/n]: n', '',
        '----------Amend Entries-----------', 'Save final item? [y/N]: y',
        ' - INFO -  Attempting to remove OphydItem (name=tst_base_pim) from the collection ...',
        ' - INFO -  Storing item OphydItem (name=tst_base_pim) ...',
        ' - INFO -  Adding / Modifying information for tst_base_pim ...',
        ' - INFO -  HappiItem OphydItem (name=tst_base_pim) has been succesfully added to the database',
        ''], id="transfer_succeeding",
    )])
def test_transfer_cli(
    from_user: str,
    expected_output: list[str],
    happi_cfg: str,
    runner: CliRunner
):
    results = runner.invoke(happi_cli, ['--path', happi_cfg, 'transfer',
                            'tst_base_pim', 'OphydItem'], input=from_user)
    assert_match_expected(results, expected_output)
    client = happi.client.Client.from_config(cfg=happi_cfg)
    item = client.find_item(name="tst_base_pim")
    assert item["type"] == 'OphydItem'
    assert getattr(item, 'parent', 'DNE') == 'DNE'


@pytest.mark.parametrize("from_user, expected_output", [
    pytest.param('\n'.join(['MY:PREFIX', 'y']), [
        'Attempting to transfer tst_minimal to OphydItem...',
        '+---------------+---------------+',
        '|  tst_minimal  |   OphydItem   |',
        '+---------------+---------------+',
        '|     active    |     active    |',
        '|      args     |      args     |',
        '|  device_class |  device_class |',
        '| documentation | documentation |',
        '|     kwargs    |     kwargs    |',
        '|      name     |      name     |',
        '|       -       |     prefix    |',
        '+---------------+---------------+',
        '',
        '----------Prepare Entries-----------',
        'Enter value for prefix, enforce=str: MY:PREFIX',
        '',
        '----------Amend Entries-----------',
        'Save final item? [y/N]: y',
        ' - INFO -  Attempting to remove OphydItem (name=tst_minimal) from the collection ...',
        ' - INFO -  Storing item OphydItem (name=tst_minimal) ...',
        ' - INFO -  Adding / Modifying information for tst_minimal ...',
        ' - INFO -  HappiItem OphydItem (name=tst_minimal) has been succesfully added to the database',
        ''], id="transfer_succeeding",
    )])
def test_transfer_cli_more(
    from_user: str,
    expected_output: list[str],
    happi_cfg: str,
    runner: CliRunner
):
    results = runner.invoke(happi_cli, ['--path', happi_cfg, 'transfer',
                            'tst_minimal', 'OphydItem'], input=from_user)
    assert_match_expected(results, expected_output)
    client = happi.client.Client.from_config(cfg=happi_cfg)
    item = client.find_item(name="tst_minimal")
    assert item["type"] == 'OphydItem'
    assert item["prefix"] == 'MY:PREFIX'


def arg_variants(variants: tuple[tuple[tuple[str]]]):
    """
    Collapse argument variants into all possible combinations.
    """
    for idx, arg_set in enumerate(itertools.product(*variants), 1):
        item = functools.reduce(
            lambda x, y: x + y,
            arg_set,
        )
        summary = f"args{idx}_" + ",".join(item)
        yield pytest.param(item, id=summary)


benchmark_arg_variants = (
    (('--duration', '0.1'), ('--iterations', '5')),
    (('--wait-connected',), ()),
    (('--tracebacks',), ()),
    (('--sort-key', 'name'), ()),
    (('--glob', '*pim'), ('--regex', '.*pim'), ()),
)


@pytest.mark.parametrize(
    "args", tuple(arg_variants(benchmark_arg_variants))
)
def test_benchmark_cli(runner: CliRunner, happi_cfg: str, args: tuple[str]):
    # Make sure the benchmark can complete in some form with valid inputs
    if "--iterations" not in args:
        # Keep the number of overall iterations reasonable
        args = (*args, "--iterations", "100")

    result = runner.invoke(
        happi_cli,
        ['--path', happi_cfg, 'benchmark'] + list(args),
    )
    assert_in_expected(
        result,
        'Benchmark completed successfully'
    )


profile_arg_variants = (
    (('--database',), ('--import',), ('--object', ),
     ('--all', ), ('--import', '--object')),
    (('--profiler', 'pcdsutils'), ('--profiler', 'cprofile'), ()),
    (('--glob', '*pim'), ('--regex', '.*pim'), ()),
)


@pytest.mark.parametrize(
    "args", tuple(arg_variants(profile_arg_variants))
)
def test_profile_cli(runner: CliRunner, happi_cfg: str, args: tuple[str]):
    # Make sure the profile can complete in some form with valid inputs
    if "pcdsutils" in args:
        if not test_line_prof:
            pytest.skip("Missing pcdsutils or line_profiler.")

    if test_line_prof:
        print("Resetting the line profiler...")
        pcdsutils.profile.reset_profiler()

    result = runner.invoke(
        happi_cli,
        ['--path', happi_cfg, 'profile'] + list(args),
    )

    assert_in_expected(
        result,
        'Profile completed successfully'
    )



import logging
import sys
import tempfile
from typing import Any
from unittest.mock import patch

import pytest
import simplejson
from click.testing import CliRunner

import happi.cli
from happi import Client, EntryInfo, HappiItem, OphydItem
from happi.backends.json_db import JSONBackend

logger = logging.getLogger(__name__)

# Conditional import of pymongo, mongomock
try:
    from mongomock import MongoClient

    from happi.backends.mongo_db import MongoBackend
    supported_backends = ['json', 'mongo']
except ImportError as exc:
    logger.warning('Error importing mongomock : -> %s', exc)
    supported_backends = ['json']

requires_mongo = pytest.mark.skipif('mongo' not in supported_backends,
                                    reason='Missing mongo')

# Conditional import of psdm_qs_cli
try:
    from psdm_qs_cli import QuestionnaireClient

    from happi.backends.qs_db import QSBackend
    has_qs_cli = True
except ImportError as exc:
    logger.warning('Error importing psdm_qs_cli : -> %s', exc)
    has_qs_cli = False


requires_questionnaire = pytest.mark.skipif(not has_qs_cli,
                                            reason='Missing psdm_qs_cli')


try:
    import pcdsdevices  # noqa
    has_pcdsdevices = True
except ImportError as exc:
    logger.warning('error importing pcdsdevices : -> %s', exc)
    has_pcdsdevices = False


requires_pcdsdevices = pytest.mark.skipif(not has_pcdsdevices,
                                          reason='Missing pcdsdevices')


requires_py39 = pytest.mark.skipif(
    sys.version_info < (3, 9),
    reason='Optional dependency needs at least py39',
)


@pytest.fixture(scope='function')
def item_info() -> dict[str, Any]:
    return {'name': 'alias',
            'z': 400,
            '_id': 'alias',
            'prefix': 'BASE:PV',
            'beamline': 'LCLS',
            'type': 'OphydItem',
            'device_class': 'types.SimpleNamespace',
            'args': list(),
            'kwargs': {'hi': 'oh hello'},
            'location_group': 'LOC',
            'functional_group': 'FUNC',
            }


@pytest.fixture(scope='function')
def item(item_info: dict[str, Any]) -> OphydItem:
    return OphydItem(**item_info)


class JinjaItem(OphydItem):
    blank_list = EntryInfo('a list', enforce=list, default=[1, 2, 3])
    blank_str = EntryInfo('a string', enforce=str, default='blank')
    blank_bool = EntryInfo('a bool', enforce=bool, default=True)
    blank_exclude = EntryInfo('omitted if default',
                              default='default',
                              include_default_as_kwarg=False)
    blank_none = EntryInfo('default is None')


@pytest.fixture(scope='function')
def item_info_jinja() -> dict[str, Any]:
    return {'name': 'alias',
            'z': 400,
            '_id': 'alias',
            'prefix': 'BASE:PV',
            'beamline': 'LCLS',
            'type': 'OphydItem',
            'device_class': 'types.SimpleNamespace',
            'args': list(),
            'kwargs': {
                'hi': 'oh hello',
                'loc': '{{location_group}}',
                'blank_list': '{{blank_list}}',
                'blank_str': '{{blank_str}}',
                'blank_bool': '{{blank_bool}}',
                'blank_none': '{{blank_none}}',
                'blank_exclude': '{{blank_exclude}}',
                'blank': '{{blank}}'
            },
            'location_group': 'LOC',
            'functional_group': 'FUNC',
            'blank_list': [1, 2, 3],
            'blank_str': 'blank',
            'blank_bool': True,
            'blank_none': None,
            'blank_exclude': 'default',
            'blank': None
            }


@pytest.fixture(scope='function')
def item_jinja(item_info_jinja: dict[str, Any]) -> OphydItem:
    return JinjaItem(**item_info_jinja)


@pytest.fixture(scope='function')
def valve_info() -> dict[str, Any]:
    return {'name': 'name',
            'z': 300,
            'prefix': 'BASE:VGC:PV',
            '_id': 'name',
            'beamline': 'LCLS',
            'mps': 'MPS:VGC:PV',
            'location_group': 'LOC',
            'functional_group': 'FUNC',
            }


@pytest.fixture(scope='function')
def valve(valve_info: dict[str, Any]) -> OphydItem:
    return OphydItem(**valve_info)


@pytest.fixture(scope='function')
def Item1():
    class Item1(HappiItem):
        dupe = EntryInfo('duplicate', enforce=int, default=1)
        one = EntryInfo('two', enforce=['one', 'zero'])
        prefix = EntryInfo('z', enforce=bool)
        # cannot be cast into Item2.bad_dupe1
        bad_dupe1 = EntryInfo('bad enforce', enforce=str)
        bad_dupe2 = EntryInfo('bad enforce', enforce=int)
        excl1_1 = EntryInfo('exclusive to Item1 #1', default='e1_1',
                            enforce=str)
        excl1_2 = EntryInfo('exclusive to Item1 #2', enforce=str)

    return Item1


@pytest.fixture(scope='function')
def item1_dev(Item1):
    return Item1(name='i1', documentation='Created as Item1',
                 one='one', bad_dupe1='hello', bad_dupe2=33)


@pytest.fixture(scope='function')
def Item2():
    class Item2(HappiItem):
        dupe = EntryInfo('duplicate', enforce=int, default=1)
        two = EntryInfo('two', enforce=['zero', 'two'], optional=False)
        bad_dupe1 = EntryInfo('bad enforce', enforce=bool)
        # cannot be cast into Item1.bad_dupe2
        bad_dupe2 = EntryInfo('bad enforce', enforce=str)
        excl2_1 = EntryInfo('exclusive to Item2 #1', default=21, enforce=int)
        excl2_2 = EntryInfo('exclusive to Item2 #2', enforce=int)

    return Item2


@pytest.fixture(scope='function')
def item2_dev(Item2):
    return Item2(name='i2', documentaiton='Created as Item2',
                 two='two', bad_dupe1=False, bad_dupe2='hallo')


@pytest.fixture(scope='function')
def mockjsonclient(item_info: dict[str, Any]):
    # Write underlying database
    with tempfile.NamedTemporaryFile(mode='w') as handle:
        simplejson.dump({item_info['name']: item_info},
                        handle)
        handle.flush()  # flush buffer to write file
        # Return handle name
        db = JSONBackend(handle.name)
        yield Client(database=db)
        # tempfile will be deleted once context manager is resolved


@pytest.fixture(scope='function')
@requires_mongo
def mockmongoclient(item_info: dict[str, Any]):
    with patch('happi.backends.mongo_db.MongoClient') as mock_mongo:
        mc = MongoClient()
        mc['test_db'].create_collection('test_collect')
        mock_mongo.return_value = mc
        # Client
        backend = MongoBackend(db='test_db', pw='test_pw', user='user', host='host',
                               collection='test_collect')
        client = Client(database=backend)
        # Insert a single device
        client.backend._collection.insert_one(item_info)
        return client


if 'mongo' in supported_backends:
    @pytest.fixture(scope='function', params=supported_backends)
    def happi_client(request, mockmongoclient, mockjsonclient):
        if request.param == 'json':
            return mockjsonclient
        if request.param == 'mongo':
            return mockmongoclient
else:
    @pytest.fixture(scope='function', params=supported_backends)
    def happi_client(request, mockjsonclient):
        return mockjsonclient


@pytest.fixture(scope='module')
def mockqsbackend():
    # Create a very basic mock class
    class MockQuestionnaireClient(QuestionnaireClient):

        def getProposalsListForRun(self, run):
            return {
                'X534': {'Instrument': 'TST', 'proposal_id': 'X534'},
                'LR32': {'Instrument': 'TST', 'proposal_id': 'LR32'},
                'LU34': {'Instrument': 'MFX', 'proposal_id': 'LU34'},
            }

        def getProposalDetailsForRun(self, run_no, proposal):
            return {
                'pcdssetup-motors-1-location': 'Hutch-main experimental',
                'pcdssetup-motors-1-name': 'sam_x',
                'pcdssetup-motors-1-purpose': 'sample x motion',
                'pcdssetup-motors-1-pvbase': 'TST:USR:MMS:01',
                'pcdssetup-motors-1-stageidentity': 'IMS MD23',
                'pcdssetup-motors-2-location': 'Hutch-main experimental',
                'pcdssetup-motors-2-name': 'sam_z',
                'pcdssetup-motors-2-purpose': 'sample z motion',
                'pcdssetup-motors-2-pvbase': 'TST:USR:MMS:02',
                'pcdssetup-motors-2-stageidentity': 'IMS MD23',
                'pcdssetup-motors-3-location': 'Hutch-main experimental',
                'pcdssetup-motors-3-name': 'sam_y',
                'pcdssetup-motors-3-purpose': 'sample y motion',
                'pcdssetup-motors-3-pvbase': 'TST:USR:MMS:03',
                'pcdssetup-motors-3-stageidentity': 'IMS MD32',
                'pcdssetup-motors-4-location': 'Hutch-main experimental',
                'pcdssetup-motors-4-name': 'sam_r',
                'pcdssetup-motors-4-purpose': 'sample rotation',
                'pcdssetup-motors-4-pvbase': 'TST:USR:MMS:04',
                'pcdssetup-motors-4-stageidentity': 'IMS MD23',
                'pcdssetup-motors-5-location': 'Hutch-main experimental',
                'pcdssetup-motors-5-name': 'sam_az',
                'pcdssetup-motors-5-purpose': 'sample azimuth',
                'pcdssetup-motors-5-pvbase': 'TST:USR:MMS:05',
                'pcdssetup-motors-5-stageidentity': 'IMS MD23',
                'pcdssetup-motors-6-location': 'Hutch-main experimental',
                'pcdssetup-motors-6-name': 'sam_flip',
                'pcdssetup-motors-6-purpose': 'sample flip',
                'pcdssetup-motors-6-pvbase': 'TST:USR:MMS:06',
                'pcdssetup-motors-6-stageidentity': 'IMS MD23',
                'pcdssetup-trig-1-delay': '0.00089',
                'pcdssetup-trig-1-eventcode': '198',
                'pcdssetup-trig-1-name': 'Overview_trig',
                'pcdssetup-trig-1-polarity': 'positive',
                'pcdssetup-trig-1-purpose': 'Overview',
                'pcdssetup-trig-1-pvbase': 'MFX:REC:EVR:02:TRIG1',
                'pcdssetup-trig-1-width': '0.00075',
                'pcdssetup-trig-2-delay': '0.000894348',
                'pcdssetup-trig-2-eventcode': '198',
                'pcdssetup-trig-2-name': 'Meniscus_trig',
                'pcdssetup-trig-2-polarity': 'positive',
                'pcdssetup-trig-2-purpose': 'Meniscus',
                'pcdssetup-trig-2-pvbase': 'MFX:REC:EVR:02:TRIG3',
                'pcdssetup-trig-2-width': '0.0005',
                'pcdssetup-ao-1-device': 'Acromag IP231 16-bit',
                'pcdssetup-ao-1-name': 'irLed',
                'pcdssetup-ao-1-purpose': 'IR LED',
                'pcdssetup-ao-2-channel': '6',
                'pcdssetup-ao-2-device': 'Acromag IP231 16-bit',
                'pcdssetup-ao-2-name': 'laser_shutter_opo',
                'pcdssetup-ao-2-purpose': 'OPO Shutter',
                'pcdssetup-ao-3-channel': '7',
                'pcdssetup-ao-3-device': 'Acromag IP231 16-bit',
                'pcdssetup-ao-3-name': 'laser_shutter_evo1',
                'pcdssetup-ao-3-purpose': 'EVO Shutter1',
                'pcdssetup-ao-4-channel': '2',
                'pcdssetup-ao-4-device': 'Acromag IP231 16-bit',
                'pcdssetup-ao-4-name': 'laser_shutter_evo2',
                'pcdssetup-ao-4-purpose': 'EVO Shutter2',
                'pcdssetup-ao-5-channel': '3',
                'pcdssetup-ao-5-device': 'Acromag IP231 16-bit',
                'pcdssetup-ao-5-name': 'laser_shutter_evo3',
                'pcdssetup-ao-5-purpose': 'EVO Shutter3',
                'pcdssetup-ao-1-pvbase': 'MFX:USR:ao1',
                'pcdssetup-ao-2-pvbase': 'MFX:USR:ao1',
                'pcdssetup-ao-3-pvbase': 'MFX:USR:ao1',
                'pcdssetup-ao-4-pvbase': 'MFX:USR:ao1',
                'pcdssetup-ao-5-pvbase': 'MFX:USR:ao1',
                'pcdssetup-ai-1-device': 'Acromag IP231 16-bit',
                'pcdssetup-ai-1-name': 'irLed',
                'pcdssetup-ai-1-purpose': 'IR LED',
                'pcdssetup-ai-1-pvbase': 'MFX:USR:ai1',
                'pcdssetup-ai-1-channel': '7',
                'pcdssetup-motors-11-purpose': 'Von Hamos vertical',
                'pcdssetup-motors-11-stageidentity': 'Beckhoff',
                'pcdssetup-motors-11-location': 'XPP goniometer',
                'pcdssetup-motors-11-pvbase': 'HXX:VON_HAMOS:MMS:01',
                'pcdssetup-motors-11-name': 'vh_y',
            }

        def getExpName2URAWIProposalIDs(self):
            return {
                'tstx53416': 'X534',
                'tstlr3216': 'LR32',
                'mfxlu3417': 'LU34',
            }

    with patch('happi.backends.qs_db.QuestionnaireClient') as qs_cli:
        # Replace QuestionnaireClient with our test version
        mock_qs = MockQuestionnaireClient(use_kerberos=False,
                                          user='user', pw='pw')
        qs_cli.return_value = mock_qs
        # Instantiate a fake device
        backend = QSBackend('tstlr3216')
        return backend


@pytest.fixture(scope='function')
def three_valves():
    valve1 = {'name': 'valve1',
              'z': 300,
              'prefix': 'BASE:VGC1:PV',
              '_id': 'VALVE1',
              'beamline': 'LCLS',
              'mps': 'MPS:VGC:PV',
              'type': 'OphydItem',
              'location_group': 'LOC',
              'functional_group': 'FUNC',
              'device_class': 'types.SimpleNamespace',
              'args': list(),
              'kwargs': {'hi': 'oh hello'},
              }

    valve2 = {'name': 'valve2',
              'z': 301,
              'prefix': 'BASE:VGC2:PV',
              '_id': 'VALVE2',
              'beamline': 'LCLS',
              'mps': 'MPS:VGC:PV',
              'type': 'OphydItem',
              'location_group': 'LOC',
              'functional_group': 'FUNC',
              'device_class': 'types.SimpleNamespace',
              'args': list(),
              'kwargs': {'hi': 'oh hello'},
              }

    valve3 = {'name': 'valve3',
              'z': 301,
              'prefix': 'BASE:VGC3:PV',
              '_id': 'VALVE3',
              'beamline': 'LCLS',
              'mps': 'MPS:VGC:PV',
              'location_group': 'LOC',
              'functional_group': 'FUNC',
              'type': 'OphydItem',
              'device_class': 'types.SimpleNamespace',
              'args': list(),
              'kwargs': {'hi': 'oh hello'},
              }

    valves = dict(
        VALVE1=valve1,
        VALVE2=valve2,
        VALVE3=valve3,
    )

    return valves


@pytest.fixture(scope='function')
def client_with_three_valves(happi_client, three_valves):
    for dev in happi_client.all_items:
        happi_client.backend.delete(dev['_id'])

    for name, valve in three_valves.items():
        happi_client.backend.save(name, valve, insert=True)
    return happi_client


@pytest.fixture(scope='function')
def db(tmp_path):
    json_path = tmp_path / 'db.json'
    json_path.write_text("""\
{
    "tst_base_pim": {
        "_id": "tst_base_pim",
        "active": true,
        "beamline": "TST",
        "creation": "Tue Jan 29 09:46:00 2019",
        "device_class": "types.SimpleNamespace",
        "last_edit": "Thu Apr 12 14:40:08 2018",
        "macros": null,
        "name": "tst_base_pim",
        "parent": null,
        "prefix": "TST:BASE:PIM",
        "screen": null,
        "stand": "BAS",
        "system": "diagnostic",
        "type": "HappiItem",
        "z": 3.0,
        "y": 40.0
    },
    "tst_base_pim2": {
        "_id": "tst_base_pim2",
        "active": true,
        "beamline": "TST",
        "creation": "Wed Jan 30 09:46:00 2019",
        "device_class": "types.SimpleNamespace",
        "kwargs": {
            "name": "{{name}}"
        },
        "last_edit": "Fri Apr 13 14:40:08 2018",
        "macros": null,
        "name": "tst_base_pim2",
        "parent": null,
        "prefix": "TST:BASE:PIM2",
        "screen": null,
        "stand": "BAS",
        "system": "diagnostic",
        "type": "HappiItem",
        "z": 6.0,
        "y": 10.0
    },
    "tst_minimal": {
        "_id": "tst_minimal",
        "creation": "Tue Jan 29 09:46:00 2019",
        "device_class": "types.SimpleNamespace",
        "last_edit": "Thu Sep 14 14:40:08 2018",
        "name": "tst_minimal",
        "type": "HappiItem"
    }
}
""")
    return str(json_path.absolute())


@pytest.fixture(scope='function')
def happi_cfg(tmp_path, db):
    happi_cfg_path = tmp_path / 'happi.cfg'
    happi_cfg_path.write_text(f"""\
[DEFAULT]'
backend=json
path={db}
""")
    return str(happi_cfg_path.absolute())


@pytest.fixture(scope='function')
def bad_db(tmp_path):
    json_path = tmp_path / 'db.json'
    json_path.write_text("""\
{
    "tst_id": {
        "_id": "tst_id",
        "active": true,
        "creation": "Tue Jan 29 09:46:00 2019",
        "device_class": "types.SimpleNamespace",
        "last_edit": "Thu Apr 12 14:40:08 2018",
        "name": "tst_name",
        "type": "HappiItem"
    },
    "tst_inst": {
        "_id": "tst_inst",
        "active": true,
        "creation": "Tue Jan 29 09:46:00 2019",
        "device_class": "invalidclassname",
        "last_edit": "Thu Apr 12 14:40:08 2018",
        "name": "tst_inst",
        "type": "HappiItem"
    },
    "tst_extra_info": {
        "_id": "tst_extra_info",
        "active": true,
        "creation": "Tue Jan 29 09:46:00 2019",
        "device_class": "types.SimpleNamespace",
        "extra_info": "extra_info_here",
        "last_edit": "Thu Apr 12 14:40:08 2018",
        "name": "tst_extra_info",
        "type": "HappiItem"
    },
    "tst_arg_mismatch": {
        "_id": "tst_arg_mismatch",
        "args": ["{{active}}"],
        "creation": "Tue Jan 29 09:46:00 2019",
        "device_class": "types.SimpleNamespace",
        "last_edit": "Thu Apr 12 14:40:08 2018",
        "name": "tst_arg_mismatch",
        "type": "HappiItem"
    },
    "tst_kwarg_mismatch": {
        "_id": "tst_kwarg_mismatch",
        "active": true,
        "args": ["{{active}}"],
        "device_class": "types.SimpleNamespace",
        "kwargs": {"creation": "{{creation}}"},
        "name": "tst_kwarg_mismatch",
        "last_edit": "Thu Apr 12 14:40:08 2018",
        "type": "HappiItem"
    },
    "tst_missing_mandatory": {
        "_id": "tst_missing_mandatory",
        "active": true,
        "args": ["{{active}}"],
        "creation": "Tue Jan 29 09:46:00 2019",
        "device_class": "types.SimpleNamespace",
        "kwargs": {"creation": "{{creation}}"},
        "name": "tst_missing_mandatory",
        "last_edit": "Thu Apr 12 14:40:08 2018",
        "type": "OphydItem"
    }
}
""")
    return str(json_path.absolute())


@pytest.fixture(scope='function')
def bad_happi_cfg(tmp_path, bad_db):
    happi_cfg_path = tmp_path / 'happi.cfg'
    happi_cfg_path.write_text(f"""\
[DEFAULT]'
backend=json
path={bad_db}
""")
    return str(happi_cfg_path.absolute())


@pytest.fixture(scope='function')
def runner():
    return CliRunner()


@pytest.fixture(autouse=True)
def skip_cleanup(monkeypatch):
    """ Monkeypatch happi_cli to skip ophyd cleanup """
    def no_op(*args, **kwargs):
        return

    monkeypatch.setattr(happi.cli, 'ophyd_cleanup', no_op)
    monkeypatch.setattr(happi.cli, 'pyepics_cleanup', no_op)



__all__ = ['backend', 'BACKENDS', 'DEFAULT_BACKEND']
import logging
import os

logger = logging.getLogger(__name__)


def _get_backend(backend):
    if backend == 'mongodb':
        from .mongo_db import MongoBackend
        return MongoBackend
    if backend == 'json':
        from .json_db import JSONBackend
        return JSONBackend
    if backend == 'qs':
        from .qs_db import QSBackend
        return QSBackend
    if backend == 'multi':
        from .multi_db import MultiBackend
        return MultiBackend
    raise ValueError(f'Unknown backend {backend!r}')


def _get_backends():
    # A hook for entrypoints or something similar later
    backends = {}
    try:
        backends['json'] = _get_backend('json')
    except ImportError as ex:
        logger.debug('JSON backend unavailable: %s', ex)

    try:
        backends['mongodb'] = _get_backend('mongodb')
    except ImportError as ex:
        logger.debug('MongoDB backend unavailable: %s', ex)

    try:
        backends['qs'] = _get_backend('qs')
    except ImportError as ex:
        logger.debug('Questionnaire backend unavailable: %s', ex)

    try:
        backends['multi'] = _get_backend('multi')
    except ImportError as ex:
        logger.debug('Multi backend unavailable: %s', ex)

    return backends


BACKENDS = _get_backends()
try:
    # Check to see if the user has specified a specific backend as an
    # environment variable. Import this as the standard backend for other
    # places in the module. A user can always override this by explicitly
    # importing the backend
    DEFAULT_BACKEND = BACKENDS[os.environ.get("HAPPI_BACKEND", 'json').lower()]
except KeyError:
    raise ImportError("Improper specification of happi backend. "
                      "Check the `$HAPPI_BACKEND` environment variable.")

# back-compatibility
backend = DEFAULT_BACKEND



"""
Core classes for Qt-based display GUIs.
"""
import sys
from typing import Any, ClassVar

from qtpy.uic import loadUiType

from ..utils import HAPPI_SOURCE_PATH


class DesignerDisplay:
    """Helper class for loading designer .ui files and adding logic."""
    filename: ClassVar[str]
    ui_form: ClassVar[Any]

    def __init_subclass__(cls):
        """Read the file when the class is created"""
        super().__init_subclass__()
        cls.ui_form, _ = loadUiType(
            str(HAPPI_SOURCE_PATH / "qt" / cls.filename)
        )

    def __init__(self, *args, **kwargs):
        """Apply the file to this widget when the instance is created"""
        super().__init__(*args, **kwargs)
        self.ui_form.setupUi(self, self)

    def retranslateUi(self, *args, **kwargs):
        """Required function for setupUi to work in __init__"""
        self.ui_form.retranslateUi(self, *args, **kwargs)

    def show_type_hints(self, fp=sys.stdout):
        """
        Show type hints of widgets included in the display for development
        purposes.

        Parameters
        ----------
        fp : file-like object, optional
            Defaults to ``sys.stdout``.
        """
        cls_attrs = set()
        obj_attrs = set(dir(self))
        annotated = set(self.__annotations__)
        for cls in type(self).mro():
            cls_attrs |= set(dir(cls))
        likely_from_ui = obj_attrs - cls_attrs - annotated
        for attr in sorted(likely_from_ui):
            try:
                obj = getattr(self, attr, None)
            except Exception:
                ...
            else:
                if obj is None:
                    continue

                print(
                    f"{attr}: {obj.__class__.__module__}"
                    f".{obj.__class__.__name__}",
                    file=fp
                )



"""
Backend implementation that combines multiple backends.
"""
import functools
import logging
import re
from typing import Any, Optional, Union

from .core import ItemMeta, ItemMetaGen, _Backend

logger = logging.getLogger(__name__)


def prevent_duplicate_ids(fn):
    """
    Decorator that remembers which document _id's have been yielded and
    prevents subsequent documents with matching _id's from being yielded

    Expects the wrapped function to yield ItemMeta documents
    """
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        unique_ids = set()
        for doc in fn(*args, **kwargs):
            if doc['_id'] not in unique_ids:
                unique_ids.add(doc['_id'])
                yield doc

    return wrapper


class MultiBackend(_Backend):
    """
    Multi-backend backend.

    Combines multiple backends, prioritizing database files in order
    of appearance.

    Parameters
    ----------
    backends : List[_Backend]
        A list of instantiated backends in order of priority.
        The backend at index 0 will have the highest priority.
    """
    def __init__(self, backends: list[_Backend]):
        self.backends = backends

    @property
    def all_items(self) -> list[ItemMeta]:
        """
        List of all items in all backends.

        In the case of duplicate items, the item in the first backend
        takes priority.

        Returns
        -------
        List[ItemMeta]
            A list of all metadata documents
        """
        items = []
        unique_ids = set()
        for bknd in self.backends:
            curr_items = bknd.all_items
            for doc in curr_items:
                if doc['_id'] not in unique_ids:
                    items.append(doc)
                    unique_ids.add(doc['_id'])

        return items

    def clear_cache(self) -> None:
        """
        Request to clear any cached data.

        Clears the cache of each backend
        """
        for bknd in self.backends:
            bknd.clear_cache()

    @prevent_duplicate_ids
    def find(self, *args, multiples=False, **kwargs) -> ItemMetaGen:
        """
        Find an instance or instances that matches the search criteria.

        Parameters
        ----------
        multiples : bool
            Find a single result or all results matching the provided
            information.
        kwargs
            Requested information.

        Yields
        ------
        Dict[str, Any]
            A generator of metadata documents
        """
        for bknd in self.backends:
            yield from bknd.find(*args, **kwargs)

    def save(self, _id, post, insert=True):
        """The current implementation of this backend is read-only."""
        raise NotImplementedError("The Multi-backend backend is read-only")

    def delete(self, _id):
        """The current implementation of this backend is read-only."""
        raise NotImplementedError("The Multi-backend backend is read-only")

    def get_by_id(self, id_: str) -> Optional[ItemMeta]:
        """
        Get an document by ID if it exists, or return None.

        Returns the first match, in config priority

        Parameters
        ----------
        id_ : str
            id to search for

        Returns
        -------
        Dict[str, Any] or None
            The requested metadata document
        """
        for bknd in self.backends:
            doc = bknd.get_by_id(id_)
            if doc:
                return doc

        return

    @prevent_duplicate_ids
    def find_range(
        self,
        key: str,
        *,
        start: Union[int, float],
        stop: Optional[Union[int, float]] = None,
        to_match: dict[str, Any]
    ) -> ItemMetaGen:
        """
        Find instances that match the search criteria, such that
        ``start <= entry[key] < stop``.

        Reports the document in the highest priority backend in the
        case of duplicates

        Parameters
        ----------
        key : str
            The database key to search
        start : Union[int, float]
            Inclusive minimum value to filter ``key`` on.
        to_match : Dict[str, Any]
            Exclusive maximum value to filter ``key`` on.
        stop : Union[int, float], optional
            Requested information, where the values must match exactly.

        Yields
        ------
        Dict[str, Any]
            metadata documents matching the provided range
        """
        for bknd in self.backends:
            yield from bknd.find_range(key, start=start, stop=stop,
                                       to_match=to_match)

    @prevent_duplicate_ids
    def find_regex(
        self,
        to_match: dict[str, Any],
        *,
        flags=re.IGNORECASE
    ) -> ItemMetaGen:
        """
        Find an instance or instances that matches the search criteria,
        using regular expressions.

        Parameters
        ----------
        to_match : dict
            Requested information, where the values are regular expressions.
        flags : RegexFlag
            flags to modify regex pattern parsing

        Yields
        ------
        Dict[str, Any]
            matching metadata documents
        """
        for bknd in self.backends:
            yield from bknd.find_regex(to_match, flags=flags)



import sys
import time
from collections import defaultdict
from pathlib import Path

# find the pre-release directory and release notes file
THIS_DIR = Path(__file__).resolve().parent
PRE_RELEASE = THIS_DIR / 'source' / 'upcoming_release_notes'
TEMPLATE = PRE_RELEASE / 'template-short.rst'
RELEASE_NOTES = THIS_DIR / 'source' / 'releases.rst'

# Set up underline constants
TITLE_UNDER = '#'
RELEASE_UNDER = '='
SECTION_UNDER = '-'


def parse_pre_release_file(path):
    """
    Return dict mapping of release notes section to lines.

    Uses empty list when no info was entered for the section.
    """
    print(f'Checking {path} for release notes.')
    with path.open('r') as fd:
        lines = fd.readlines()

    section_dict = defaultdict(list)
    prev = None
    section = None

    for line in lines:
        if prev is not None:
            if line.startswith(SECTION_UNDER * 2):
                section = prev.strip()
                continue
            if section is not None and line[0] in ' -':
                notes = section_dict[section]
                if len(line) > 6:
                    notes.append(line)
                section_dict[section] = notes
        prev = line

    return section_dict


def extend_release_notes(path, version, release_notes):
    """
    Given dict mapping of section to lines, extend the release notes file.
    """
    with path.open('r') as fd:
        lines = fd.readlines()

    new_lines = ['\n', '\n']
    date = time.strftime('%Y-%m-%d')
    release_title = f'{version} ({date})'
    new_lines.append(release_title + '\n')
    new_lines.append(len(release_title) * RELEASE_UNDER + '\n')
    new_lines.append('\n')
    for section, section_lines in release_notes.items():
        if section == 'Contributors':
            section_lines = sorted(list(set(section_lines)))
        if len(section_lines) > 0:
            new_lines.append(section + '\n')
            new_lines.append(SECTION_UNDER * len(section) + '\n')
            new_lines.extend(section_lines)
            new_lines.append('\n')

    output_lines = lines[:2] + new_lines + lines[2:]

    print('Writing out release notes file')
    for line in new_lines:
        print(line.strip('\n'))
    with path.open('w') as fd:
        fd.writelines(output_lines)


def main(version_number: str):
    section_notes = parse_pre_release_file(TEMPLATE)
    to_delete = []
    for path in PRE_RELEASE.iterdir():
        if path.name[0] in '1234567890':
            to_delete.append(path)
            extra_notes = parse_pre_release_file(path)
            for section, notes in section_notes.items():
                notes.extend(extra_notes[section])
                section_notes[section] = notes

    extend_release_notes(RELEASE_NOTES, version_number, section_notes)

    print(
        "* Wrote release notes.  Please perform the following manually:",
        file=sys.stderr,
    )
    for path in to_delete:
        print(f" git rm {path}", file=sys.stderr)
    print(f" git add {RELEASE_NOTES}", file=sys.stderr)


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} VERSION_NUMBER", file=sys.stderr)
        sys.exit(1)

    version_number = sys.argv[1]

    if not version_number.startswith("v"):
        print(
            f"Version number should start with 'v': {version_number}",
            file=sys.stderr
        )
        sys.exit(1)

    main(version_number)



from typing import Any

import pytest
from click.testing import CliRunner

from happi.prompt import enforce_list, read_user_dict
from happi.utils import EnforceError


def test_user_dict(runner: CliRunner):
    default_dict = {'default_key': 'default_value'}

    # normal operation
    with runner.isolation('key1\nvalue1'):
        result = read_user_dict('prompt', default=default_dict)

    assert result == {'key1': 'value1'}

    # read default
    with runner.isolation('\n'):
        result = read_user_dict('prompt', default=default_dict)

    assert result == default_dict

    # reject keywords
    with runner.isolation('is\nnotis\n1\n'):
        result = read_user_dict('prompt', default=default_dict)

    assert result == {'notis': 1}

    # replace values
    with runner.isolation('key\n1\nkey\n2'):
        result = read_user_dict('prompt', default=default_dict)

    assert result == {'key': 2}


@pytest.mark.parametrize('user_in', (
    ['a', 'b', 2, 3],
    "['a', 'b', 2, 3]"
))
def test_enforce_list(user_in: Any):
    result = enforce_list(user_in)
    assert result == ['a', 'b', 2, 3]


@pytest.mark.parametrize('user_in', (
    'a',
    "['a', 'b'=2, 2, 3]",
    '[1,2,3,4.5.4]'
))
def test_enforce_list_fail(user_in: str):
    with pytest.raises(EnforceError):
        _ = enforce_list(user_in)



import copy
import io
import re
from typing import Any

import pytest

from ..client import Client
from ..errors import ContainerError
from ..item import EntryInfo, HappiItem, OphydItem


def test_get(item: OphydItem, item_info: dict[str, Any]):
    assert item.name == item_info['name']


def test_init(item: OphydItem, item_info: dict[str, Any]):
    assert item.prefix == item_info['prefix']
    assert item.name == item_info['name']


def test_list_enforce():
    # Generic HappiItem with a list enforce
    class MyDevice(HappiItem):
        list_attr = EntryInfo(enforce=['a', 'b', 'c'],
                              enforce_doc='list only')

    # Make sure we can set without error
    d = MyDevice()
    d.list_attr = 'b'
    assert d.list_attr == 'b'
    # Mase sure we can not set outside the list
    d = MyDevice()
    with pytest.raises(ValueError) as excinfo:
        d.list_attr = 'd'

    assert 'list only' in str(excinfo)


def test_regex_enforce():
    class MyDevice(HappiItem):
        re_attr = EntryInfo(enforce=re.compile(r'[A-Z]{2}$'),
                            enforce_doc='only 2 chars')

    d = MyDevice()
    d.re_attr = 'AB'

    d = MyDevice()
    with pytest.raises(ValueError) as excinfo:
        d.re_attr = 'ABC'

    assert 'only 2 chars' in str(excinfo)


@pytest.mark.parametrize('type_spec, value, expected',
                         [(int, 0, 0), (int, 1, 1), (int, 2.0, 2),
                          (str, 'hat', 'hat'), (str, 5, '5'),
                          (bool, True, True), (bool, 0, False),
                          (bool, 'true', True), (bool, 'NO', False)])
def test_type_enforce_ok(type_spec: type, value: Any, expected: Any):
    entry = EntryInfo(enforce=type_spec)
    assert entry.enforce_value(value) == expected


@pytest.mark.parametrize('type_spec, value',
                         [(int, 'cats'),
                          (bool, '24'), (bool, 'catastrophe')])
def test_type_enforce_exceptions(type_spec: type, value: Any):
    entry = EntryInfo(enforce=type_spec, enforce_doc='bad type')
    with pytest.raises(ValueError) as excinfo:
        entry.enforce_value(value)

    assert 'bad type' in str(excinfo)


def test_set(item: OphydItem):
    item.name = 'new_name'
    assert item.name == 'new_name'


def test_optional(item: OphydItem):
    assert item.documentation is None


def test_enforce(item: OphydItem):
    with pytest.raises(ValueError):
        item.name = 'Invalid!Name'


def test_container_error():
    with pytest.raises(ContainerError):
        class MyDevice(HappiItem):
            fault = EntryInfo(enforce=int, default='not-int')


def test_mandatory_info(item: OphydItem):
    for info in ('prefix', 'name'):
        assert info in item.mandatory_info


def test_restricted_attr():
    with pytest.raises(TypeError):
        class MyDevice(HappiItem):
            info_names = EntryInfo()


def test_post(item: OphydItem, item_info: dict[str, Any]):
    post = item.post()
    assert post['prefix'] == item_info['prefix']
    assert post['name'] == item_info['name']


def test_show_info(item: OphydItem, item_info: dict[str, Any]):
    f = io.StringIO()
    item.show_info(handle=f)
    f.seek(0)
    out = f.read()
    item_info.pop('_id')
    assert '_id' not in out
    assert all([info in out for info in item_info.keys()])


def test_device_equivalance():
    a = HappiItem(name='abcd', prefix='b')
    b = HappiItem(name='abcd', prefix='b')
    c = HappiItem(name='cbcd', prefix='b')
    assert a == b
    assert not c == a


def test_dictify():
    a = HappiItem(name='abcd', prefix='b')
    assert dict(a) == a.post()


def test_device_copy():
    a = HappiItem(name='abcd', prefix='b')
    b = copy.copy(a)
    assert dict(a) == dict(b)


def test_device_deepcopy():
    a = HappiItem(name='abcd', prefix='abc', kwargs={'abc': 'def'})

    c = copy.deepcopy(a)
    assert a.kwargs == c.kwargs
    assert id(a.kwargs) != id(c.kwargs)


def test_add_and_save(valve: OphydItem, happi_client: Client):
    valve.active = True
    happi_client.add_item(valve)
    valve.active = False
    valve.save()

    assert not happi_client[valve.name].item.active



"""
Backend implemenation using the ``simplejson`` package.
"""
import contextlib
import logging
import math
import os
import os.path
import re
import shutil
import uuid
from typing import Any, Callable, Optional, Union

import simplejson as json

from .. import utils
from ..errors import DuplicateError, SearchError
from .core import ItemMeta, ItemMetaGen, _Backend

logger = logging.getLogger(__name__)

# A sentinel for keys that are missing for comparisons below.
_MISSING = object()


@contextlib.contextmanager
def _load_and_store_context(backend):
    """Context manager used to load, and optionally store the JSON database."""
    db = backend._load_or_initialize()
    yield db
    backend.store(db)


class JSONBackend(_Backend):
    """
    JSON database.

    The happi information is kept in a single large dictionary that is stored
    using the ``simplejson`` package.

    Parameters
    ----------
    path : str
        Path to JSON file.
    initialize : bool, optional
        Initialize a new empty JSON file to begin filling.
    cfg_path : str, optional
        Path to the happi config.
    """

    def __init__(
        self,
        path: str,
        initialize: bool = False,
        cfg_path: Optional[str] = None
    ) -> None:
        self._load_cache: dict[str, ItemMeta] = None
        # Determine the cfg dir and build path to json db based on that unless we're initted w/o a config
        if cfg_path is not None:
            cfg_dir = os.path.dirname(cfg_path)
            self.path = utils.build_abs_path(cfg_dir, path)
        else:
            self.path = path
        # Create a new JSON file if requested
        if initialize:
            self.initialize()

    def clear_cache(self) -> None:
        """Clear the loaded cache."""
        self._load_cache = None

    def _load_or_initialize(self) -> Optional[dict[str, ItemMeta]]:
        """Load an existing database or initialize a new one."""
        if self._load_cache is None:
            try:
                self._load_cache = self.load()
            except FileNotFoundError:
                logger.debug("Initializing new database")
                self.initialize()
                self._load_cache = self.load()

        return self._load_cache

    @property
    def all_items(self) -> list[ItemMeta]:
        """All of the items in the database."""
        json = self._load_or_initialize()
        return list(json.values())

    def initialize(self):
        """
        Initialize a new JSON file database.

        Raises
        ------
        PermissionError
            If the JSON file specified by ``path`` already exists.

        Notes
        -----
        This exists because the `.store` and `.load` methods assume that the
        given path already points to a readable JSON file. In order to begin
        filling a new database, an empty but valid JSON file is created.
        """

        # Do not overwrite existing databases
        if os.path.exists(self.path) and os.stat(self.path).st_size > 0:
            raise PermissionError("File {} already exists. Can not initialize "
                                  "a new database.".format(self.path))
        # Dump an empty dictionary
        self.store({})

    def load(self) -> dict[str, ItemMeta]:
        """Load the JSON database."""
        with open(self.path) as f:
            raw_json = f.read()

        # Allow for empty files to be considered valid databases:
        return json.loads(raw_json) if raw_json else {}

    def store(self, db: dict[str, ItemMeta]) -> None:
        """
        Stash the database in the JSON file.

        This is a two-step process:

        1. Write the database out to a temporary file
        2. Move the temporary file over the previous database.

        Step 2 is an atomic operation, ensuring that the database
        does not get corrupted by an interrupted json.dump.

        Parameters
        ----------
        db : dict
            Dictionary to store in JSON.
        """
        temp_path = self._temp_path()
        try:
            with open(temp_path, 'w') as fd:
                json.dump(db, fd, sort_keys=True, indent=4)

            if os.path.exists(self.path):
                shutil.copymode(self.path, temp_path)
            shutil.move(temp_path, self.path)
        except BaseException as ex:
            logger.debug('JSON db move failed: %s', ex, exc_info=ex)
            # remove temporary file
            if os.path.exists(temp_path):
                os.remove(temp_path)
            raise

    def _temp_path(self) -> str:
        """
        Return a temporary path to write the json file to during "store".

        Includes a hash for uniqueness
        (in the cases where multiple temp files are written at once).
        """
        directory = os.path.dirname(self.path)
        filename = (
            f"_{str(uuid.uuid4())[:8]}"
            f"_{os.path.basename(self.path)}"
        )
        return os.path.join(directory, filename)

    def _iterative_compare(self, comparison: Callable) -> ItemMetaGen:
        """
        Yields documents in which ``comparison(name, doc)`` returns `True`.

        Parameters
        ----------
        comparison : Callable
            A comparison function with a signature of (item_id, doc).
        """

        db = self._load_or_initialize()
        if not db:
            return

        for name, doc in db.items():
            try:
                if comparison(name, doc):
                    yield doc
            except Exception as ex:
                logger.debug('Comparison method failed: %s', ex, exc_info=ex)

    def get_by_id(self, id_: str) -> ItemMeta:
        """Get an item by ID if it exists, or return None."""
        db = self._load_or_initialize()
        return db.get(id_)

    def find(self, to_match: dict[str, Any]) -> ItemMetaGen:
        """
        Find an instance or instances that matches the search criteria.

        Parameters
        ----------
        to_match : dict
            Requested information, all of which must match.
        """

        def comparison(name, doc):
            return all(
                value == doc.get(key, _MISSING)
                for key, value in to_match.items()
            )

        yield from self._iterative_compare(comparison)

    def find_range(
        self,
        key: str,
        *,
        start: Union[int, float],
        stop: Optional[Union[int, float]] = None,
        to_match: dict[str, Any]
    ) -> ItemMetaGen:
        """
        Find an instance or instances that matches the search criteria, such
        that ``start <= entry[key] < stop``.

        Parameters
        ----------
        key : str
            The database key to search.
        start : int or float
            Inclusive minimum value to filter ``key`` on.
        end : float, optional
            Exclusive maximum value to filter ``key`` on.
        to_match : dict
            Requested information, where the values must match exactly.
        """

        def comparison(name, doc):
            if all(value == doc.get(k, _MISSING)
                   for k, value in to_match.items()):
                try:
                    return start <= doc[key] < stop
                except Exception:
                    ...
            return False

        if key in to_match:
            raise ValueError('Cannot specify the same key in `to_match` as '
                             'the key for the range.')
        if stop is None:
            stop = math.inf
        if start >= stop:
            raise ValueError(f"Invalid range: {start} >= {stop}")

        yield from self._iterative_compare(comparison)

    def find_regex(
        self,
        to_match: dict[str, Any],
        *,
        flags=re.IGNORECASE
    ) -> ItemMetaGen:
        """
        Find an instance or instances that matches the search criteria,
        using regular expressions.

        Parameters
        ----------
        to_match : dict
            Requested information, where the values are regular expressions.
        """

        def comparison(name, doc):
            return regexes and all(key in doc and regex.match(str(doc[key]))
                                   for key, regex in regexes.items())

        regexes = {
            key: re.compile(value, flags=flags)
            for key, value in to_match.items()
        }

        yield from self._iterative_compare(comparison)

    def save(
        self,
        _id: str,
        post: dict[str, Any],
        insert: bool = True
    ) -> None:
        """
        Save information to the database.

        Parameters
        ----------
        _id : str
            ID of item.
        post : dict
            Information to place in database.
        insert : bool, optional
            Whether or not this is a new item to the database.

        Raises
        ------
        DuplicateError
            If ``insert`` is `True`, but there is already an item with the
            provided ``_id``.
        SearchError
            If ``insert`` is `False`, but there is no item with the provided
            ``_id``.
        PermissionError
            If the write operation fails due to issues with permissions.
        """

        with _load_and_store_context(self) as db:
            # New item
            if insert:
                if _id in db.keys():
                    raise DuplicateError(f"Item {_id} already exists")
                # Add _id keyword
                post.update({'_id': _id})
                # Add to database
                db[_id] = post
            # Updating item
            else:
                # Edit information
                try:
                    db[_id].update(post)
                except KeyError:
                    raise SearchError(f"No item found {_id}")

    def delete(self, _id: str) -> None:
        """
        Delete an item instance from the database.

        Parameters
        ----------
        _id : str
            ID of item.

        Raises
        ------
        PermissionError
            If the write operation fails due to issues with permissions.
        """

        with _load_and_store_context(self) as db:
            try:
                db.pop(_id)
            except KeyError:
                raise SearchError(
                    f'ID not found in database: {_id!r}'
                ) from None
